// [[[$article >= 11]]]
/* Copyright (C) 2005  Thomas Petazzoni

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.
*/

#include <sos/types.h>
#include <sos/klibc.h>
#include <sos/kmalloc.h>
#include <sos/list.h>

#include "netdev_generic.h"
#include "net_protos.h"

struct arp {
  sos_ui16_t hardware_type; /* 0x1 */
  sos_ui16_t protocol_type; /* 0x0800 */
  sos_ui8_t  hardware_size; /* 6 */
  sos_ui8_t  protocol_size; /* 4 */
  sos_ui16_t opcode;
#define ARP_OPCODE_REQUEST 1
#define ARP_OPCODE_REPLY   2

  sos_ui8_t  srcmac[6];
  sos_ui32_t srcip;
  sos_ui8_t  dstmac[6];
  sos_ui32_t dstip;
}__attribute__((packed));

struct arp_entry {
  sos_ui32_t ip;
  sos_ui8_t  mac[6];
  sos_bool_t valid;
  struct arp_waiting_pkt *waiting_pkts;
  struct arp_entry *prev, *next;
};

struct arp_waiting_pkt {
  struct sos_netdev *dev;
  struct sos_netdev_pkt *pkt;
  sos_ui16_t eth_type;
  struct arp_waiting_pkt *prev, *next;
};

struct arp_entry *arp_cache = NULL;

static sos_ret_t sos_arp_wait_entry (struct arp_entry *entry, struct sos_netdev *dev,
				     struct sos_netdev_pkt *pkt, sos_ui16_t eth_type)
{
  struct arp_waiting_pkt *wait_pkt;

  wait_pkt = (struct arp_waiting_pkt *) sos_kmalloc (sizeof(struct arp_waiting_pkt), 0);
  if (! wait_pkt)
    return -SOS_ENOMEM;

  wait_pkt->dev = dev;
  wait_pkt->pkt = pkt;
  wait_pkt->eth_type = eth_type;

  list_add_tail (entry->waiting_pkts, wait_pkt);

  return SOS_OK;
}

static void sos_arp_wakeup_entry (struct arp_entry *entry)
{
  struct arp_waiting_pkt *wait_pkt;

  while (! list_is_empty (entry->waiting_pkts))
    {
      wait_pkt = list_pop_head (entry->waiting_pkts);
      sos_eth_send_packet (wait_pkt->dev, wait_pkt->pkt, entry->mac, wait_pkt->eth_type);
      sos_kfree ((sos_vaddr_t) wait_pkt);
    }
}

static sos_ret_t send_arp_packet (struct sos_netdev * dev,
				  sos_ui16_t opcode,
				  sos_ui32_t srcip, const sos_ui8_t *srcmac,
				  sos_ui32_t dstip, const sos_ui8_t *dstmac)
{
  struct sos_netdev_pkt *pkt;
  struct arp      *arp_pkt;

  pkt = sos_eth_alloc_pkt (dev,
			   sizeof (struct arp),
			   TRUE, 0);

  arp_pkt = (struct arp *) & pkt->contents[pkt->pdu_offset];

  memset(arp_pkt, 0, sizeof(struct arp));

  arp_pkt->hardware_type = htons(0x1);
  arp_pkt->protocol_type = htons(0x800);
  arp_pkt->hardware_size = 6;
  arp_pkt->protocol_size = 4;
  arp_pkt->opcode        = htons(opcode);
  memcpy (arp_pkt->srcmac, srcmac, 6);
  arp_pkt->srcip = htonl (srcip);
  memcpy (arp_pkt->dstmac, dstmac, 6);
  arp_pkt->dstip = htonl (dstip);

  return sos_eth_send_packet (dev, pkt, dstmac, 0x0806);
}


sos_ret_t sos_arp_recv_packet (struct sos_netdev * dev,
			       struct sos_netdev_pkt * pkt)
{
  struct arp *arp_pkt;

  arp_pkt = (struct arp *) & pkt->contents[pkt->pdu_offset];

  if (0 == memcmp(& arp_pkt->dstip, dev->netaddr, 4))
    {
      if (ntohs(arp_pkt->opcode) == ARP_OPCODE_REQUEST)
	{
	  sos_ui32_t myip;
	  memcpy(& myip, dev->netaddr, 4);

	  sos_bochs_printf("[ARP] Request DSTMAC=0x%x 0x%x 0x%x 0x%x 0x%x 0x%x\n",
			   arp_pkt->srcmac[0], arp_pkt->srcmac[1], arp_pkt->srcmac[2],
			   arp_pkt->srcmac[3], arp_pkt->srcmac[4], arp_pkt->srcmac[5]);
	  send_arp_packet (dev, ARP_OPCODE_REPLY,
			   ntohl(myip), (sos_ui8_t*)& dev->hwaddr,
			   ntohl(arp_pkt->srcip), arp_pkt->srcmac);
	}
      else if (ntohs(arp_pkt->opcode) == ARP_OPCODE_REPLY)
	{
	  struct arp_entry *entry;
	  int n;

	  sos_bochs_printf ("[ARP] Reply telling that 0x%ld is at 0x%x:0x%x:0x%x:0x%x:0x%x:0x%x\n",
			    arp_pkt->srcip,
			    arp_pkt->srcmac[0], arp_pkt->srcmac[1], arp_pkt->srcmac[2],
			    arp_pkt->srcmac[3], arp_pkt->srcmac[4], arp_pkt->srcmac[5]);

	  list_foreach (arp_cache, entry, n)
	    {
	      if (entry->ip == ntohl(arp_pkt->srcip) && ! entry->valid)
		{
		  memcpy (entry->mac, arp_pkt->srcmac, sizeof (entry->mac));
		  entry->valid = TRUE;
		  sos_arp_wakeup_entry (entry);
		}
	    }

	}
    }

  return SOS_OK;
}

sos_ret_t sos_arp_send_packet (struct sos_netdev * dev,
			       struct sos_netdev_pkt *pkt,
			       sos_ui32_t dst_ip,
			       sos_ui16_t eth_type)
{
  struct arp_entry *entry;
  int n, ret;
  sos_ui8_t bcastmac[6] = { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff };

  list_foreach (arp_cache, entry, n)
    {
      if (entry->ip == dst_ip)
	{
	  if (entry->valid)
	    {
	      sos_bochs_printf ("[ARP] Valid entry found for 0x%lx\n", dst_ip);
	      return sos_eth_send_packet (dev, pkt, entry->mac, eth_type);
	    }
	  else
	    {
	      sos_bochs_printf ("[ARP] Invalid entry found for 0x%lx\n", dst_ip);
	      return sos_arp_wait_entry (entry, dev, pkt, eth_type);
	    }
	}
    }

  sos_bochs_printf ("[ARP] No entry found for 0x%lx\n", dst_ip);

  /* No ARP entry for the destination IP */
  entry = (struct arp_entry *) sos_kmalloc (sizeof(struct arp_waiting_pkt), 0);
  if (! entry)
    return -SOS_ENOMEM;

  entry->ip = dst_ip;
  memset (entry->mac, 0, sizeof(entry->mac));
  entry->valid = FALSE;
  entry->waiting_pkts = NULL;

  list_add_tail (arp_cache, entry);

  ret = sos_arp_wait_entry (entry, dev, pkt, eth_type);
  if (ret != SOS_OK)
    return ret;

  return send_arp_packet (dev, ARP_OPCODE_REQUEST, NETADDR_2_IPV4(dev->netaddr), dev->hwaddr,
			  dst_ip, bcastmac);
}
// [[[]]]
