// [[[$article >= 10]]]
/* Copyright (C) 2006 David Decotigny, Thomas Petazzoni

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.
*/

#include <sos/kmalloc.h>
#include <sos/klibc.h>
#include <sos/assert.h>
#include <sos/list.h>
#include <sos/ksynch.h>
#include <sos/uaccess.h>
#include <sos/physmem.h>

#include <sos/fs.h>
#include <sos/fs_nscache.h>
#include <sos/fs_pagecache.h>

/**
 * @file fs_ext2.c
 *
 * Implementation of the ext2 filesystem, initially designed by Rémy
 * Card, and which was the default filesystem for the GNU/Linux
 * operating system for a long time (it is now replaced by its
 * journalized version, ext3).
 *
 * For more information about ext2 :
 *  - http://en.wikipedia.org/wiki/Ext2
 *  - http://www.nongnu.org./ext2-doc/ext2.html
 *  - http://uranus.it.swin.edu.au/~jn/explore2fs/es2fs.htm
 */

/** The description of the "Ext2 FS" */
static struct sos_fs_manager_type ext2fs_type;

/**
 * An on-disk ext2 superblock
 */
struct ext2_sb
{
  unsigned long  s_inodes_count;
  unsigned long  s_blocks_count;
  unsigned long  s_r_blocks_count;
  unsigned long  s_free_block_count;
  unsigned long  s_free_inodes_count;
  unsigned long  s_first_data_block;
  unsigned long  s_log_block_size;
  long           s_log_frag_size;
  unsigned long  s_blocks_per_group;
  unsigned long  s_frags_per_group;
  unsigned long  s_inodes_per_group;
  unsigned long  s_mtime;
  unsigned long  s_wtime;
  unsigned short s_mnt_count;
  unsigned short s_max_mnt_count;
  unsigned short s_magic;
  unsigned short s_tate;
  unsigned short s_errors;
  unsigned short s_minor_rev_level;
  unsigned long  s_lastcheck;
  unsigned long  s_checkinterval;
  unsigned long  s_creator_os;
  unsigned long  s_rev_level;
  unsigned short s_def_resuid;
  unsigned short s_def_resgid;
  /* EXT2_DYNAMIC_REV */
  unsigned long  s_first_ino;
  unsigned short s_inode_size;
  unsigned short s_block_group_nr;
  unsigned long  s_feature_compat;
  unsigned long  s_feature_incompat;
  unsigned long  s_feature_ro_compat;
  unsigned long  s_uuid[4];
  unsigned char  s_name[16];
  unsigned char  s_last_mounted[64];
  unsigned long  s_algo_bitmap;
  unsigned long  unused[205];
};

/** Location of the first ext2 superblock in the device */
#define EXT2_SB_OFFSET 1024

/** Ext2 magic */
#define EXT2_MAGIC 0xEF53

/** Ext2 root directory inode number */
#define EXT2_ROOT_INODE 0x2

struct ext2_group_desc
{
  unsigned long  bg_block_bitmap;
  unsigned long  bg_inode_bitmap;
  unsigned long  bg_inode_table;
  unsigned short bg_free_blocks_count;
  unsigned short bg_free_inodes_count;
  unsigned short bg_used_dirs_count;
  unsigned short bd_pad;
  unsigned short bg_reserved[6];
};

struct ext2_inode
{
#define EXT2_S_IFMT(mode) (((mode) & 0xF000))
#define EXT2_S_IFSOCK	  0xC000
#define EXT2_S_IFLNK	  0xA000
#define EXT2_S_IFREG	  0x8000
#define EXT2_S_IFBLK	  0x6000
#define EXT2_S_IFDIR	  0x4000
#define EXT2_S_IFCHR	  0x2000
#define EXT2_S_IFIFO	  0x1000
#define EXT2_S_IRUSR	  0x0100
#define EXT2_S_IWUSR	  0x0080
#define EXT2_S_IXUSR	  0x0040
  unsigned short i_mode;
  unsigned short i_uid;
  unsigned long i_size;
  unsigned long i_atime;
  unsigned long i_ctime;
  unsigned long i_mtime;
  unsigned long i_dtime;
  unsigned short i_gid;
  unsigned short i_links_count;
  unsigned long i_blocks;
  unsigned long i_flags;
  unsigned long i_osd1;
#define EXT2_N_BLOCKS    15
#define EXT2_NDIR_BLOCKS 12 /**< Number of direct blocks */
#define EXT2_IND_BLOCK   12 /**< Id of the indirect block */
#define EXT2_DIND_BLOCK  13 /**< Id of the double-indirect block */
#define EXT2_TIND_BLOCK  14 /**< Id of the triple-indirect block */
  unsigned long i_block[EXT2_N_BLOCKS];
  unsigned long i_generation;
  unsigned long i_file_acl;
  unsigned long i_dir_acl;
  unsigned long i_faddr;
  unsigned long i_osd2[3];
};

struct ext2_dir_entry
{
  unsigned long inode;
  unsigned short rec_len;
  unsigned char name_len;
  unsigned char file_type;
  unsigned char name[0];
};

#define EXT2_FT_UNKNOWN	 0
#define EXT2_FT_REG_FILE 1
#define EXT2_FT_DIR	 2
#define EXT2_FT_CHRDEV	 3
#define EXT2_FT_BLKDEV	 4
#define EXT2_FT_FIFO	 5
#define EXT2_FT_SOCK	 6
#define EXT2_FT_SYMLINK	 7
#define EXT2_FT_MAX	 8


/**
 * Structure of a FS file or dir for ext2fs.
 *
 * Structural inheritance of sos_fs_node: "super" is the ancestor (in
 * "OOP" terms). The "super" field might not be the first field: from
 * a sos_fs_node, the corresponding ext2fs_node is given by the
 * sos_fs_node::custom_data field
 */
struct ext2fs_node
{
  /** The ancestor */
  struct sos_fs_node super;

  /** List of blocks */
  unsigned long blocks[EXT2_N_BLOCKS];

  /** Size */
  sos_size_t size;

  union
  {
    /* A file */
    struct
    {
      /** Page cache */
      struct sos_fs_pagecache *pc;

      /** Map resource, in order to be able to mmap() the file */
      struct sos_umem_vmm_mapped_resource mapres;

    } file;

  }; /* Anonymous union (gcc extension) */
};


/**
 * A ext2fs FS instance
 *
 * Structural inheritance of sos_fs_manager_instance: "super" is the
 * ancestor (in "OOP" terms). The "super" field might not be the first
 * field: from a sos_fs_manager_instance, the corresponding
 * ext2fs_instance is given by the
 * sos_fs_manager_instance::custom_data field
 */
struct ext2fs_instance
{
  /** Ancestor */
  struct sos_fs_manager_instance super;

  struct sos_kmutex lock;

  struct sos_blockdev_instance * device;

  unsigned long blocks_count;
  unsigned long block_size;
  unsigned long blocks_per_group;
  unsigned long inodes_per_group;
};

static sos_ret_t ext2fs_find_block(struct ext2fs_node *node,
				   struct ext2fs_instance *fs,
				   unsigned long log_block,
				   unsigned long *phy_block)
{

  /* Number of blocks referenced by a block */
#define EXT2_BLOCKS_ENTRIES_NB(fs) ((fs)->block_size / sizeof(unsigned int))

  /* After the blocks covered by the double indirect block */
  if (log_block >= (EXT2_NDIR_BLOCKS + (EXT2_BLOCKS_ENTRIES_NB(fs) *
					EXT2_BLOCKS_ENTRIES_NB(fs))))
      {
	sos_bochs_printf ("Oups, triple-indirect blocks are not supported!\n");
	return -SOS_EIO;
      }


  if (log_block >= (EXT2_NDIR_BLOCKS + EXT2_BLOCKS_ENTRIES_NB(fs)))
      {
	unsigned long id_in_indir_blk, id_in_dir_blk;
	unsigned int dir_blk, blk;
	sos_size_t read_len = sizeof(unsigned long);
	sos_ret_t retval;

	log_block -= (EXT2_NDIR_BLOCKS + EXT2_BLOCKS_ENTRIES_NB(fs));
	id_in_indir_blk = log_block / EXT2_BLOCKS_ENTRIES_NB(fs);
	id_in_dir_blk   = log_block % EXT2_BLOCKS_ENTRIES_NB(fs);

	retval = sos_blockdev_kernel_read(fs->device,
					  node->blocks[13] * fs->block_size + id_in_indir_blk * 4,
					  (sos_vaddr_t) & dir_blk, & read_len);
	if (SOS_OK != retval)
	  return retval;
	if (read_len != sizeof(unsigned int))
	  return -SOS_EIO;

	retval = sos_blockdev_kernel_read(fs->device,
					  dir_blk * fs->block_size + id_in_dir_blk * 4,
					  (sos_vaddr_t) & blk, & read_len);

	if (SOS_OK != retval)
	  return retval;
	if (read_len != sizeof(unsigned int))
	  return -SOS_EIO;

	*phy_block = blk;

	return SOS_OK;
      }

  if (log_block >= EXT2_NDIR_BLOCKS)
    {
      unsigned int blk;
      sos_size_t read_len = sizeof(unsigned long);
      sos_ret_t retval;

      log_block -= EXT2_NDIR_BLOCKS;

      retval = sos_blockdev_kernel_read(fs->device,
					node->blocks[12] * fs->block_size + log_block * 4,
					(sos_vaddr_t) & blk, & read_len);
      if (SOS_OK != retval)
	return retval;
      if (read_len != sizeof(unsigned int))
	return -SOS_EIO;

      *phy_block = blk;

      return SOS_OK;
    }

  *phy_block = node->blocks[log_block];

  return SOS_OK;
}

static sos_ret_t ext2fs_inode_read (struct ext2fs_instance *fs,
				    unsigned long inode_id,
				    struct ext2_inode *inode)
{
  unsigned long group, index;
  struct ext2_group_desc group_desc;
  sos_ret_t ret;
  sos_size_t len;

  group = (inode_id - 1) / fs->inodes_per_group;
  index = (inode_id - 1) % fs->inodes_per_group;

  len = sizeof(struct ext2_group_desc);
  ret = sos_blockdev_kernel_read (fs->device, sizeof(struct ext2_sb) + 1024 + group * sizeof(struct ext2_group_desc),
				  (sos_vaddr_t) & group_desc, &len);
  if (ret || len != sizeof(struct ext2_group_desc))
    return ret;

  len = sizeof(struct ext2_inode);
  ret = sos_blockdev_kernel_read (fs->device,
				  (group_desc.bg_inode_table * fs->block_size +
				   index * sizeof(struct ext2_inode)),
				  (sos_vaddr_t) inode, &len);
  if (ret || len != sizeof(struct ext2_inode))
    return ret;

  sos_bochs_printf ("[Inode %lu: %lu/%lu] ctime=%lu, block0=%lu\n",
		    inode_id, group, index, inode->i_ctime, inode->i_block[0]);

  return SOS_OK;
}

static sos_ret_t ext2fs_contents_read (struct ext2fs_instance *fs,
				       struct ext2fs_node *ext2fsnode,
				       sos_lsoffset_t position,
				       sos_genaddr_t dest_addr,
				       sos_size_t *to_read,
				       sos_bool_t use_pgcache)
{
  sos_uoffset_t offset;
  sos_size_t remaining;
  sos_ret_t retval;

  remaining = *to_read;

  offset = 0;
  while (remaining > 0)
    {
      sos_uoffset_t block    = position / fs->block_size;
      sos_uoffset_t off_block = position % fs->block_size;
      unsigned long phy_block;
      sos_size_t sz_block;
      sos_size_t read_len;

      if (use_pgcache)
	{
	  read_len = remaining;
	  retval = sos_fs_pagecache_read(ext2fsnode->file.pc, position,
					 dest_addr, & read_len);

	  sos_bochs_printf("read through pagecache: %d, %d -> %d\n",
			   retval, remaining, read_len);

	  if (retval != -SOS_ENOENT)
	    {
	      remaining -= read_len;
	      offset    += read_len;
	      position  += read_len;

	      if (retval == SOS_OK)
		continue;
	      else
		break;
	    }
	}

      if (off_block + remaining > fs->block_size)
	sz_block = fs->block_size - off_block;
      else
	sz_block = remaining;

      retval = ext2fs_find_block (ext2fsnode, fs, block, & phy_block);
      if (SOS_OK != retval)
	break;

#if 0
      sos_bochs_printf ("ext2fs_read(), reading %d at offset %lu, blk %d->%lu, off_blk %d\n",
			sz_block, (unsigned long) offset,
			block, phy_block, off_block);
#endif

      read_len = sz_block;
      if (dest_addr.is_user)
	retval   = sos_blockdev_user_read(fs->device,
					  phy_block * fs->block_size + off_block,
					  dest_addr.addr + offset, & read_len);
      else
	retval   = sos_blockdev_kernel_read(fs->device,
					    phy_block * fs->block_size + off_block,
					    dest_addr.addr + offset, & read_len);

      position       += read_len;
      offset         += read_len;
      remaining      -= read_len;

      if (SOS_OK != retval)
	{
	  sos_bochs_printf("ext2fs_read(), read failed with %d\n", retval);
	  break;
	}

      if (read_len != sz_block)
	{
	  sos_bochs_printf("ext2fs_read(), incomplete read (%d != %d)\n",
			   read_len, sz_block);
	  break;
	}
    }

  *to_read = offset;

  return SOS_OK;
}

static sos_ret_t ext2fs_contents_write (struct ext2fs_instance *fs,
					struct ext2fs_node *ext2fsnode,
					sos_lsoffset_t position,
					sos_genaddr_t src_addr,
					sos_size_t *to_write,
					sos_bool_t bypass_pgcache)
{
  sos_uoffset_t offset;
  sos_size_t remaining;
  sos_ret_t retval;

  SOS_ASSERT_FATAL(ext2fsnode->super.type == SOS_FS_NODE_REGULAR_FILE);

  remaining = *to_write;

  offset = 0;
  while (remaining > 0)
    {
      sos_uoffset_t block    = position / fs->block_size;
      sos_uoffset_t off_block = position % fs->block_size;
      unsigned long phy_block;
      sos_size_t sz_block;
      sos_size_t write_len;


      if (! bypass_pgcache)
	{
	  write_len = remaining;
	  retval = sos_fs_pagecache_write(ext2fsnode->file.pc, position,
					  src_addr, & write_len,
					  FALSE);

	  sos_bochs_printf("write through pagecache: %d, %d -> %d\n",
			   retval, remaining, write_len);

	  if (retval != -SOS_ENOENT)
	    {
	      remaining -= write_len;
	      offset    += write_len;
	      position  += write_len;

	      if (retval == SOS_OK)
		continue;
	      else
		break;
	    }
	}

      if (off_block + remaining > fs->block_size)
	sz_block = fs->block_size - off_block;
      else
	sz_block = remaining;

      retval = ext2fs_find_block (ext2fsnode, fs, block, & phy_block);
      if (SOS_OK != retval)
	break;

#if 0
      sos_bochs_printf ("ext2fs_contents_write(), writing %d at offset %lu, blk %d->%lu, off_blk %d\n",
			sz_block, (unsigned long) offset,
			block, phy_block, off_block);
#endif

      write_len = sz_block;
      if (src_addr.is_user)
	retval   = sos_blockdev_user_write(fs->device,
					   phy_block * fs->block_size + off_block,
					   src_addr.addr + offset, & write_len);
      else
	retval   = sos_blockdev_kernel_write(fs->device,
					     phy_block * fs->block_size + off_block,
					     src_addr.addr + offset, & write_len);

      position       += write_len;
      offset         += write_len;
      remaining      -= write_len;

      if (SOS_OK != retval)
	{
	  sos_bochs_printf("ext2fs_contents_write(), write failed with %d\n", retval);
	  break;
	}

      if (write_len != sz_block)
	{
	  sos_bochs_printf("ext2fs_contents_write(), incomplete write (%d != %d)\n",
			   write_len, sz_block);
	  break;
	}
    }

  *to_write = offset;

  return SOS_OK;
}

/* ********************************************************
 * Helper functions
 */


/* Helper function to serialize "disk" accesses */
inline static void ext2fs_lock(struct ext2fs_node *a_node)
{
  struct ext2fs_instance * fs = (struct ext2fs_instance*)a_node->super.fs->custom_data;
  sos_kmutex_lock(& fs->lock, NULL);
}


/* Helper function to serialize "disk" accesses */
inline static void ext2fs_unlock(struct ext2fs_node *a_node)
{
  struct ext2fs_instance * fs = (struct ext2fs_instance*)a_node->super.fs->custom_data;
  sos_kmutex_unlock(& fs->lock);
}

/* ********************************************************
 * Opened file operations
 */


static sos_ret_t
ext2fs_duplicate_opened_file(struct sos_fs_opened_file *this,
			     const struct sos_process  * for_owner,
			     struct sos_fs_opened_file **result)
{
  *result = (struct sos_fs_opened_file*)
    sos_kmalloc(sizeof(struct sos_fs_opened_file), 0);
  if (! *result)
    return -SOS_ENOMEM;

  memcpy(*result, this, sizeof(*this));
  (*result)->owner = for_owner;
  return SOS_OK;
}


static sos_ret_t
ext2fs_seek(struct sos_fs_opened_file *this,
	    sos_lsoffset_t offset,
	    sos_seek_whence_t whence,
	    /* out */ sos_lsoffset_t * result_position)
{
  sos_lsoffset_t ref_offs;
  struct ext2fs_node * ext2fsnode;

  ext2fsnode = (struct ext2fs_node*)
    sos_fs_nscache_get_fs_node(this->direntry)->custom_data;

  if ( (ext2fsnode->super.type != SOS_FS_NODE_REGULAR_FILE)
       && (ext2fsnode->super.type != SOS_FS_NODE_SYMLINK))
    return -SOS_ENOSUP;

  *result_position = this->position;
  switch (whence)
    {
    case SOS_SEEK_SET:
      ref_offs = 0;
      break;

    case SOS_SEEK_CUR:
      ref_offs = this->position;
      break;

    case SOS_SEEK_END:
      ref_offs = ext2fsnode->size;
      break;

    default:
      return -SOS_EINVAL;
    }

  if (offset < -ref_offs)
    return -SOS_EINVAL;

  this->position = ref_offs + offset;
  *result_position = this->position;
  return SOS_OK;
}

static sos_ret_t ext2fs_read(struct sos_fs_opened_file *this,
			     sos_uaddr_t dest_buf,
			     sos_size_t * /* in/out */len)
{
  sos_ret_t retval;
  struct ext2fs_node * ext2fsnode;
  struct ext2fs_instance *fs;
  sos_size_t length;
  sos_size_t size;
  SOS_GENADDR_DECL(gaddr, TRUE, dest_buf);

  sos_bochs_printf("ext2fs_read(), size %d\n", *len);

  ext2fsnode = (struct ext2fs_node*)
    sos_fs_nscache_get_fs_node(this->direntry)->custom_data;

  fs = (struct ext2fs_instance*)
    ext2fsnode->super.fs->custom_data;

  ext2fs_lock(ext2fsnode);

  if (ext2fsnode->super.type != SOS_FS_NODE_REGULAR_FILE &&
      ext2fsnode->super.type != SOS_FS_NODE_SYMLINK)
    {
      ext2fs_unlock (ext2fsnode);
      return -SOS_ENOSUP;
    }

  size = ext2fsnode->size;

  if (this->position >= size)
    {
      sos_bochs_printf("ext2fs_read(), after the end of the file\n");
      *len = 0;
      ext2fs_unlock (ext2fsnode);
      return SOS_OK;
    }

  if (this->position + *len >= size)
    length = size - this->position;
  else
    length = *len;

  if (ext2fsnode->super.type == SOS_FS_NODE_SYMLINK)
    {
      if (size < sizeof(ext2fsnode->blocks))
	{
	  sos_vaddr_t from;
	  from = (sos_vaddr_t) (& ext2fsnode->blocks) + this->position;

	  sos_memcpy_to_user (dest_buf, from, length);

	  *len = length;
	  ext2fs_unlock(ext2fsnode);
	  return SOS_OK;
	}
    }

  retval = ext2fs_contents_read(fs, ext2fsnode, this->position,
				gaddr, & length, TRUE);

  ext2fs_unlock(ext2fsnode);

  this->position += length;
  *len            = length;

  sos_bochs_printf("Returning with size read: %d\n", *len);

  return retval;
}


static sos_ret_t ext2fs_write(struct sos_fs_opened_file *this,
			     sos_uaddr_t src_buf,
			     sos_size_t * /* in/out */len)
{
  struct ext2fs_node * ext2fsnode;
  struct ext2fs_instance * fs;
  SOS_GENADDR_DECL(gaddr, TRUE, src_buf);
  sos_size_t length;
  sos_ret_t retval;

  sos_bochs_printf ("Writing...\n");

  ext2fsnode = (struct ext2fs_node*)
    sos_fs_nscache_get_fs_node(this->direntry)->custom_data;

  fs = (struct ext2fs_instance*)
    ext2fsnode->super.fs->custom_data;

  if ( (ext2fsnode->super.type != SOS_FS_NODE_REGULAR_FILE)
       && (ext2fsnode->super.type != SOS_FS_NODE_SYMLINK))
    return -SOS_ENOSUP;

  ext2fs_lock(ext2fsnode);

  /* Cannot increase the size of the file for the moment */
  if (*len + this->position > ext2fsnode->size)
    length = ext2fsnode->size - this->position;
  else
    length = *len;

  retval = ext2fs_contents_write(fs, ext2fsnode, this->position,
				 gaddr, & length, FALSE);

  ext2fs_unlock(ext2fsnode);

  this->position += length;
  *len            = length;

  sos_fs_mark_dirty(this);

  return retval;
}


static sos_ret_t ext2fs_new_mapping(struct sos_umem_vmm_vr *vr);
static sos_ret_t ext2fs_mmap(struct sos_fs_opened_file *this,
			     sos_uaddr_t *uaddr, sos_size_t size,
			     sos_ui32_t access_rights,
			     sos_ui32_t flags,
			     sos_luoffset_t offset)
{
  struct ext2fs_node * ext2fsnode;

  sos_bochs_printf ("mmap() called on %p, size %d\n", this, size);

  ext2fsnode = (struct ext2fs_node*)
    sos_fs_nscache_get_fs_node(this->direntry)->custom_data;

  if (ext2fsnode->super.type != SOS_FS_NODE_REGULAR_FILE)
    return -SOS_ENOSUP;

  return sos_umem_vmm_map(sos_process_get_address_space(this->owner),
			  uaddr, size, access_rights,
			  flags, & ext2fsnode->file.mapres, offset);
}


static sos_ret_t ext2fs_readdir(struct sos_fs_opened_file *this,
				struct sos_fs_dirent * result)
{
  struct ext2fs_node * ext2fsnode;
  unsigned int nb;
  struct ext2fs_instance *fs;
  sos_lsoffset_t offset;

  ext2fsnode = (struct ext2fs_node*)
    sos_fs_nscache_get_fs_node(this->direntry)->custom_data;

  fs = (struct ext2fs_instance *) ext2fsnode->super.fs->custom_data;

  offset = 0;
  nb     = 0;
  while (offset < ext2fsnode->size)
    {
      struct ext2_dir_entry direntry;
      char buf[SOS_FS_DIRENT_NAME_MAXLEN];
      sos_size_t len;
      sos_ret_t retval;
      sos_fs_node_type_t type;
      SOS_GENADDR_DECL(genaddr, FALSE, (sos_ui32_t) & direntry);

      /* Read the directory entry */
      len = sizeof(struct ext2_dir_entry);
      retval = ext2fs_contents_read (fs, ext2fsnode, offset, genaddr, & len, TRUE);
      if (SOS_OK != retval)
	return retval;
      if (len != sizeof(struct ext2_dir_entry))
	return -SOS_EIO;

      if (direntry.inode == 0)
	break;

      if (direntry.file_type == EXT2_FT_REG_FILE)
	type = SOS_FS_NODE_REGULAR_FILE;
      else if (direntry.file_type == EXT2_FT_DIR)
	type = SOS_FS_NODE_DIRECTORY;
      else if (direntry.file_type == EXT2_FT_SYMLINK)
	type = SOS_FS_NODE_SYMLINK;
      else
	continue;

      if (nb == this->position)
	{
	  /* Read the name */
	  SOS_GENADDR_DECL(genbufaddr, FALSE, (sos_ui32_t) buf);
	  len = direntry.name_len;
	  retval = ext2fs_contents_read (fs, ext2fsnode,
					 (offset + sizeof(struct ext2_dir_entry)),
					 genbufaddr, & len, TRUE);
	  if (SOS_OK != retval)
	    return retval;
	  if (len != direntry.name_len)
	    return -SOS_EIO;

	  buf[direntry.name_len] = 0;

	  sos_bochs_printf("found: '%s' => %lu\n", buf,
			   direntry.inode);

	  result->storage_location = direntry.inode;
	  result->type             = type;
	  strzcpy (result->name, buf, direntry.name_len + 1);
	  result->namelen           = direntry.name_len + 1;
	  this->position++;

	  return SOS_OK;
	}

      nb++;
      offset += direntry.rec_len;
    }

  return -SOS_ENOENT;
}


static struct sos_fs_ops_opened_file ext2fs_ops_opened_file =
  (struct sos_fs_ops_opened_file){
    .seek     = ext2fs_seek,
    .read     = ext2fs_read,
    .write    = ext2fs_write,
    .mmap     = ext2fs_mmap
  };


static struct sos_fs_ops_opened_dir ext2fs_ops_opened_dir =
  (struct sos_fs_ops_opened_dir){
    .readdir  = ext2fs_readdir
  };


/* ********************************************************
 * FS node operations
 */

static sos_ret_t ext2fs_stat_node(struct sos_fs_node * this,
				  struct sos_fs_stat * result)
{
  struct ext2fs_node * ext2fsnode = (struct ext2fs_node*)this->custom_data;

  sos_bochs_printf("stat_node(), storage location=%llu\n",
		   this->storage_location);

  /* Establish the major/minor fields */
  result->st_rdev.device_class    = 0;
  result->st_rdev.device_instance = 0;

  if ( (this->type == SOS_FS_NODE_DEVICE_CHAR)
       || (this->type == SOS_FS_NODE_DEVICE_BLOCK) )
    {
      /* If the current node is a special device: get the major/minor
	 directly from it */
      result->st_rdev.device_class    = this->dev_id.device_class;
      result->st_rdev.device_instance = this->dev_id.device_instance;
    }
  else
    {
      /* Otherwise, retrieve it from the device that it is mounted
	 on (might not exist...) */
      struct sos_fs_node * rootdev = this->fs->device;
      if (rootdev)
	{
	  result->st_rdev.device_class    = rootdev->dev_id.device_class;
	  result->st_rdev.device_instance = rootdev->dev_id.device_instance;
	}
    }

  result->st_type                 = this->type;
  result->st_storage_location     = this->storage_location;
  result->st_access_rights        = this->access_rights;
  result->st_nlink                = this->ondisk_lnk_cnt;
  result->st_size                 = ext2fsnode->size;

  return SOS_OK;
}


static sos_ret_t ext2fs_truncate(struct sos_fs_node *this,
				 sos_lsoffset_t length)
{
  sos_ret_t retval;
  struct ext2fs_node * ext2fsnode;

  ext2fsnode = (struct ext2fs_node*) this->custom_data;

  if ( (ext2fsnode->super.type != SOS_FS_NODE_REGULAR_FILE)
       && (ext2fsnode->super.type != SOS_FS_NODE_SYMLINK))
    return -SOS_ENOSUP;

  ext2fs_lock(ext2fsnode);
  return -SOS_EIO;
  ext2fs_unlock(ext2fsnode);

  return retval;
}


static sos_ret_t ext2fs_sync_node(struct sos_fs_node *this)
{
  struct ext2fs_node * ext2fsnode = (struct ext2fs_node*)this->custom_data;
  sos_ret_t retval;

  sos_bochs_printf("ext2fs_sync_node[%p]\n", this);

  if (this->type != SOS_FS_NODE_REGULAR_FILE)
    return SOS_OK;

  retval = sos_fs_pagecache_sync(ext2fsnode->file.pc);

  sos_bochs_printf("end of ext2fs_sync_node[%p]\n", this);

  return retval;
}


static sos_ret_t ext2fs_chmod_node(struct sos_fs_node * this,
				   sos_ui32_t access_rights)
{
  this->access_rights = access_rights;
  return SOS_OK;
}


/** Callback when nothing (in particular no sos_fs_nscache_node) make
    reference to the node */
static sos_ret_t ext2fs_node_destructor(struct sos_fs_node * this)
{
  struct ext2fs_node * ext2fsnode = (struct ext2fs_node*)this->custom_data;

  /*
  if (this->type == SOS_FS_NODE_REGULAR_FILE)
    sos_fs_pagecache_delete_cache(ext2fsnode->file.pc);
  */

  sos_kfree((sos_vaddr_t) ext2fsnode);
  return SOS_OK;
}


static struct sos_fs_node_ops_file ext2fs_ops_file =
  (struct sos_fs_node_ops_file){
    .truncate = ext2fs_truncate,
    .stat     = ext2fs_stat_node,
    .chmod    = ext2fs_chmod_node,
    .sync     = ext2fs_sync_node
  };


static sos_ret_t ext2fs_new_opened_file(struct sos_fs_node * this,
					const struct sos_process * owner,
					sos_ui32_t open_flags,
					struct sos_fs_opened_file ** result_of)
{
  struct sos_fs_opened_file * of;

  of = (struct sos_fs_opened_file*)sos_kmalloc(sizeof(*of), 0);
  if (! of)
    return -SOS_ENOMEM;

  memset(of, 0x0, sizeof(*of));
  of->owner      = owner;
  of->duplicate  = ext2fs_duplicate_opened_file;
  of->open_flags = open_flags;
  of->ops_file   = & ext2fs_ops_opened_file;
  if (this->type == SOS_FS_NODE_DIRECTORY)
    of->ops_dir  = & ext2fs_ops_opened_dir;

  sos_bochs_printf ("new_opened_file() %p => %p\n", this, of);

  *result_of = of;
  return SOS_OK;
}


static sos_ret_t ext2fs_close_opened_file(struct sos_fs_node * this,
					  struct sos_fs_opened_file * of)
{
  sos_bochs_printf("ext2fs_close_opened_file[this=%p,of=%p]\n", this, of);
  //  sos_kfree((sos_vaddr_t)of);
  return SOS_OK;
}


static sos_ret_t ext2fs_symlink_expand(struct sos_fs_node *this,
				       char const  ** target,
				       sos_size_t * target_len)
{
  struct ext2fs_node * ext2fsnode;
  struct ext2fs_instance * fs;
  char *str;

  ext2fsnode = (struct ext2fs_node*) this->custom_data;
  fs = (struct ext2fs_instance*) ext2fsnode->super.fs->custom_data;

  sos_bochs_printf("symlink_expand()\n");

  ext2fs_lock(ext2fsnode);

  if (ext2fsnode->super.type == SOS_FS_NODE_SYMLINK)
    {
      /* Will be freed by the FS infrastructure */
      str = (char*) sos_kmalloc(ext2fsnode->size, 0);
      if (! str)
	return -SOS_ENOMEM;

      if (ext2fsnode->size < sizeof(ext2fsnode->blocks))
	memcpy (str, ext2fsnode->blocks, ext2fsnode->size);
      else
	{
	  SOS_GENADDR_DECL(genaddr, FALSE, (sos_ui32_t) str);
	  sos_size_t len;
	  sos_ret_t retval;

	  len = ext2fsnode->size;
	  retval = ext2fs_contents_read (fs, ext2fsnode, 0, genaddr, & len, TRUE);
	  if (SOS_OK != retval)
	    {
	      sos_kfree ((sos_vaddr_t) str);
	      return retval;
	    }
	  if (len != ext2fsnode->size)
	    {
	      sos_kfree ((sos_vaddr_t) str);
	      return -SOS_EIO;
	    }
	}

      *target = str;
      *target_len = ext2fsnode->size;
      ext2fs_unlock(ext2fsnode);
      return SOS_OK;
    }

  ext2fs_unlock(ext2fsnode);
  return -SOS_ENOENT;
}


static struct sos_fs_node_ops_symlink ext2fs_ops_symlink
  = (struct sos_fs_node_ops_symlink){
    .expand = ext2fs_symlink_expand
  };


static sos_ret_t ext2fs_dir_lookup(struct sos_fs_node *this,
				   const char * name, sos_ui16_t namelen,
				   sos_ui64_t * result_storage_location)
{
  struct ext2fs_node * ext2fsnode;
  struct ext2fs_instance *fs;

  sos_lsoffset_t offset;

  sos_bochs_printf ("ext2fs_dir_lookup() %p: '%s'\n", this, name);

  ext2fsnode = (struct ext2fs_node*)this->custom_data;
  fs         = (struct ext2fs_instance *) this->fs->custom_data;

  offset = 0;
  while (offset < ext2fsnode->size)
    {
      struct ext2_dir_entry direntry;
      char buf[SOS_FS_DIRENT_NAME_MAXLEN];
      sos_size_t len;
      sos_ret_t retval;
      sos_fs_node_type_t type;
      SOS_GENADDR_DECL(genaddr, FALSE, (sos_ui32_t) & direntry);
      SOS_GENADDR_DECL(genbufaddr, FALSE, (sos_ui32_t) buf);

      /* Read the directory entry */
      len = sizeof(struct ext2_dir_entry);
      retval = ext2fs_contents_read (fs, ext2fsnode, offset, genaddr, & len, TRUE);
      if (SOS_OK != retval)
	return retval;
      if (len != sizeof(struct ext2_dir_entry))
	return -SOS_EIO;

      if (direntry.inode == 0)
	break;

      if (direntry.file_type == EXT2_FT_REG_FILE)
	type = SOS_FS_NODE_REGULAR_FILE;
      else if (direntry.file_type == EXT2_FT_DIR)
	type = SOS_FS_NODE_DIRECTORY;
      else if (direntry.file_type == EXT2_FT_SYMLINK)
	type = SOS_FS_NODE_SYMLINK;
      else
	goto skip;

      if (direntry.name_len != namelen)
	goto skip;

      /* Read the name */
      len = direntry.name_len;
      retval = ext2fs_contents_read (fs, ext2fsnode,
				     (offset + sizeof(struct ext2_dir_entry)),
				     genbufaddr, & len, TRUE);
      if (SOS_OK != retval)
	return retval;
      if (len != direntry.name_len)
	return -SOS_EIO;

      buf[direntry.name_len] = 0;
      sos_bochs_printf ("direntry inode='%lu', name='%s', type='0x%x' (offset=%lld, rec_len=%d)\n",
			direntry.inode, name, direntry.file_type, offset, direntry.rec_len);

      if (! strncmp(buf, name, namelen))
	{
	  *result_storage_location = direntry.inode;
	  return SOS_OK;
	}

    skip:
      offset += direntry.rec_len;
    }

  return -SOS_ENOENT;
}


static sos_ret_t ext2fs_link(struct sos_fs_node *this,
			     const struct sos_process *actor,
			     const char * entry_name, sos_ui16_t entry_namelen,
			     struct sos_fs_node * node)
{
#if 0
  struct ext2fs_node * parent = (struct ext2fs_node*)this->custom_data;
  struct ext2fs_direntry * direntry;

  sos_bochs_printf("ext2fs_link() %p\n", this);

  direntry = (struct ext2fs_direntry*)sos_kmalloc(sizeof(*direntry), 0);
  if (! direntry)
    return -SOS_ENOMEM;

  direntry->name = (char*)sos_kmalloc(entry_namelen + 1, 0);
  if (! direntry->name)
    {
      sos_kfree((sos_vaddr_t)direntry->name);
      return -SOS_ENOMEM;
    }

  memcpy(direntry->name, entry_name, entry_namelen);
  direntry->name[entry_namelen] = '\0';

  direntry->fsnode = node;
// [[[$enclosing && $master && $debug_ref]]]
  sos_bochs_printf("## INC ONDISK REF %p %d++\n", node, node->ondisk_lnk_cnt);
// [[[]]]
  node->ondisk_lnk_cnt ++;
  this->ondisk_lnk_cnt ++;
  list_add_tail_named(parent->dir.list_entries, direntry,
		      sibling_prev, sibling_next);

  /* Update the index of the new entry in order for the next readdirs
     to fetch this new entry */
  parent->dir.top_creation_order ++;
  direntry->creation_order = parent->dir.top_creation_order;
#endif

  return -SOS_EIO;
}


static sos_ret_t
ext2fs_unlink(struct sos_fs_node *this,
	      const struct sos_process *actor,
	      const char * entry_name, sos_ui16_t entry_namelen)
{
#if 0
  struct ext2fs_node * parent = (struct ext2fs_node*)this->custom_data;
  struct ext2fs_direntry * direntry;
  int nbentries;

  list_foreach_forward_named(parent->dir.list_entries,
			     direntry, nbentries,
			     sibling_prev, sibling_next)
    {
      if (!memcmp(entry_name, direntry->name, entry_namelen)
	  && !direntry->name[entry_namelen])
	{
	  list_delete_named(parent->dir.list_entries, direntry,
			    sibling_prev, sibling_next);
// [[[$enclosing && $master && $debug_ref]]]
	  sos_bochs_printf("## DEC ONDISK REF %p %d--\n", direntry->fsnode, direntry->fsnode->ondisk_lnk_cnt);
	  SOS_ASSERT_FATAL(direntry->fsnode->ondisk_lnk_cnt > 0);
	  SOS_ASSERT_FATAL(this->ondisk_lnk_cnt);
// [[[]]]
	  direntry->fsnode->ondisk_lnk_cnt --;
	  this->ondisk_lnk_cnt --;
	  sos_kfree((sos_vaddr_t)direntry);
	  return SOS_OK;
	}
    }

  return -SOS_ENOENT;
#endif
  return -SOS_EIO;
}


static struct sos_fs_node_ops_dir ext2fs_ops_dir
  = (struct sos_fs_node_ops_dir){
    .lookup = ext2fs_dir_lookup,
    .link   = ext2fs_link,
    .unlink = ext2fs_unlink
  };


/* ********************************************************
 * FS instance operations
 */

/* Further defined */
static sos_ret_t ext2fs_sync_dirty_page(sos_luoffset_t offset,
					sos_vaddr_t dirty_page,
					void * custom_data);

/** Fetch a node from the disk, represented in ext2 by an ext2_inode
    structure. */
static sos_ret_t
ext2fs_fetch_node_from_disk(struct sos_fs_manager_instance * this,
			    sos_ui64_t storage_location,
			    struct sos_fs_node ** result)
{
  /* The "disk" is simply the ram */
  struct ext2fs_node * ext2fsnode;
  struct ext2_inode inode;
  sos_ret_t ret;

  sos_bochs_printf("##### Fetch %llu\n", storage_location);

  ret = ext2fs_inode_read ((struct ext2fs_instance*) this,
			   storage_location, & inode);
  if (SOS_OK != ret)
    {
      sos_bochs_printf("Could not fetch inode %llu (%d)\n",
		       storage_location, ret);
      return ret;
    }

  ext2fsnode = (struct ext2fs_node *)
    sos_kmalloc (sizeof(struct ext2fs_node), 0);
  if (! ext2fsnode)
    return -SOS_ENOMEM;

  memset (ext2fsnode, 0, sizeof(struct ext2fs_node));

  *result = & ext2fsnode->super;

  ext2fsnode->super.fs                = this;
  ext2fsnode->super.storage_location  = storage_location;
  ext2fsnode->super.ondisk_lnk_cnt    = inode.i_links_count;
  ext2fsnode->super.inmem_ref_cnt     = 1;
  ext2fsnode->super.destructor        = ext2fs_node_destructor;
  ext2fsnode->super.ops_file          = &ext2fs_ops_file;
  ext2fsnode->super.new_opened_file   = ext2fs_new_opened_file;
  ext2fsnode->super.close_opened_file = ext2fs_close_opened_file;
  ext2fsnode->super.custom_data       = ext2fsnode;

  memcpy (ext2fsnode->blocks, inode.i_block, sizeof(ext2fsnode->blocks));

  if (inode.i_mode & EXT2_S_IRUSR)
    ext2fsnode->super.access_rights |= SOS_FS_READABLE;
  if (inode.i_mode & EXT2_S_IWUSR)
    ext2fsnode->super.access_rights |= SOS_FS_WRITABLE;
  if (inode.i_mode & EXT2_S_IXUSR)
    ext2fsnode->super.access_rights |= SOS_FS_EXECUTABLE;

  ext2fsnode->size = inode.i_size;

  switch (EXT2_S_IFMT(inode.i_mode))
    {
    case EXT2_S_IFREG:
      ext2fsnode->super.type = SOS_FS_NODE_REGULAR_FILE;
      ext2fsnode->file.mapres.allowed_access_rights = 0;
      if (ext2fsnode->super.access_rights & SOS_FS_READABLE)
	ext2fsnode->file.mapres.allowed_access_rights |= SOS_VM_MAP_PROT_READ;
      if (ext2fsnode->super.access_rights & SOS_FS_WRITABLE)
	ext2fsnode->file.mapres.allowed_access_rights |= SOS_VM_MAP_PROT_WRITE;
      if (ext2fsnode->super.access_rights & SOS_FS_EXECUTABLE)
	ext2fsnode->file.mapres.allowed_access_rights |= SOS_VM_MAP_PROT_EXEC;
      ext2fsnode->file.mapres.custom_data = ext2fsnode;
      ext2fsnode->file.mapres.mmap        = ext2fs_new_mapping;
      break;
    case EXT2_S_IFDIR:
      ext2fsnode->super.type    = SOS_FS_NODE_DIRECTORY;
      ext2fsnode->super.ops_dir = &ext2fs_ops_dir;
      break;
    case EXT2_S_IFCHR:
      ext2fsnode->super.type = SOS_FS_NODE_DEVICE_CHAR;
      break;
    case EXT2_S_IFBLK:
      ext2fsnode->super.type = SOS_FS_NODE_DEVICE_BLOCK;
      break;
    case EXT2_S_IFLNK:
      ext2fsnode->super.type = SOS_FS_NODE_SYMLINK;
      ext2fsnode->super.ops_symlink = &ext2fs_ops_symlink;
      break;
    default:
      sos_kfree((sos_vaddr_t) ext2fsnode);
      sos_bochs_printf("Unknown inode.i_mode: 0x%x\n", EXT2_S_IFMT(inode.i_mode));
      return -SOS_ENOENT;
    }

  ext2fsnode->file.pc = sos_fs_pagecache_new_cache(ext2fs_sync_dirty_page,
						   ext2fsnode);
  if (! ext2fsnode->file.pc)
    {
      sos_kfree((sos_vaddr_t) ext2fsnode);
      return -SOS_ENOMEM;
    }

  sos_bochs_printf("##### Fetch %llu OK (type=%s) (%p))\n",
		   storage_location,
		   ((ext2fsnode->super.type == SOS_FS_NODE_REGULAR_FILE)
		    ? "file" : "dir"),
		   ext2fsnode);

  return SOS_OK;
}


static sos_ret_t
ext2fs_allocate_new_node(struct sos_fs_manager_instance * this,
			 sos_fs_node_type_t type,
			 const struct sos_process * creator,
			 sos_ui32_t access_rights,
			 sos_ui32_t flags,
			 struct sos_fs_node ** result)
{
#if 0
  struct ext2fs_node * ext2fsnode;

  /* Allow only DIRs, FILEs or special device files */
  if ((type != SOS_FS_NODE_REGULAR_FILE)
      && (type != SOS_FS_NODE_SYMLINK)
      && (type != SOS_FS_NODE_DIRECTORY)
      && (type != SOS_FS_NODE_DEVICE_CHAR)
      && (type != SOS_FS_NODE_DEVICE_BLOCK))
    return -SOS_ENOSUP;

  ext2fsnode = (struct ext2fs_node*) sos_kmalloc(sizeof(*ext2fsnode), 0);
  if (! ext2fsnode)
    return -SOS_ENOMEM;

  memset(ext2fsnode, 0x0, sizeof(*ext2fsnode));
  *result = & ext2fsnode->super;

  /* Initialize "file" */
  (*result)->inmem_ref_cnt     = 1;
  (*result)->custom_data       = ext2fsnode;
  (*result)->storage_location  = (sos_ui64_t)((sos_vaddr_t)ext2fsnode);
  (*result)->type              = type;
  (*result)->access_rights     = access_rights;
  (*result)->destructor        = ext2fs_node_destructor;
  if ((type != SOS_FS_NODE_DEVICE_CHAR)
      && (type != SOS_FS_NODE_DEVICE_BLOCK))
    {
      (*result)->new_opened_file   = ext2fs_new_opened_file;
      (*result)->close_opened_file = ext2fs_close_opened_file;
    }

  if (type == SOS_FS_NODE_SYMLINK)
    (*result)->ops_symlink = & ext2fs_ops_symlink;
  else if (type == SOS_FS_NODE_DIRECTORY)
    (*result)->ops_dir = & ext2fs_ops_dir;

  return SOS_OK;
#endif
  return -SOS_ENOMEM;
}

/* ********************************************************
 * FS type (mount/umount) operations
 */

static sos_ret_t ext2fs_mount(struct sos_fs_manager_type * this,
			      struct sos_fs_node * device,
			      const char * args,
			      struct sos_fs_manager_instance ** mounted_fs)
{
  sos_ret_t retval;
  struct ext2fs_instance * fs;
  struct sos_fs_node * fsnode_root;
  struct sos_hash_table * hash;
  struct ext2_sb *sb;
  sos_size_t len;

  *mounted_fs = (struct sos_fs_manager_instance*)NULL;

  sos_bochs_printf ("In mount\n");

  if (!device || device->type != SOS_FS_NODE_DEVICE_BLOCK)
    {
      sos_bochs_printf ("No block device specified\n");
      return -SOS_ENODEV;
    }

  if (! device->block_device)
    {
      sos_bochs_printf ("Unable to open the block device\n");
      return -SOS_ENODEV;
    }

  sos_blockdev_ref_instance(device->dev_id.device_class,
			    device->dev_id.device_instance);

  sb = (struct ext2_sb*) sos_kmalloc(sizeof(struct ext2_sb), 0);
  if (! sb)
    return -SOS_ENOMEM;

  len = sizeof(struct ext2_sb);
  retval = sos_blockdev_kernel_read (device->block_device, EXT2_SB_OFFSET,
				     (sos_vaddr_t) sb, &len);
  if (SOS_OK != retval || len != sizeof(struct ext2_sb))
    {
      sos_kfree((sos_vaddr_t) sb);
      return retval;
    }

  if (sb->s_magic != EXT2_MAGIC)
    {
      sos_kfree((sos_vaddr_t) sb);
      return -SOS_EINVAL;
    }

  sos_bochs_printf("blkdev=0x%x\n", (unsigned int) device->block_device);

  /* Create FS node hash table */
  hash = sos_hash_create("ext2fs H", struct sos_fs_node,
			 sos_hash_ui64,
			 sos_hash_key_eq_ui64, 17,
			 storage_location, hlink_nodecache);
  if (! hash)
    {
      sos_kfree((sos_vaddr_t) sb);
      return -SOS_ENOMEM;
    }

  fs = (struct ext2fs_instance*)
    sos_kmalloc(sizeof(struct ext2fs_instance), 0);
  if (! fs)
    {
      sos_kfree((sos_vaddr_t) sb);
      sos_hash_dispose(hash);
      return -SOS_ENOMEM;
    }

  memset(fs, 0x0, sizeof(struct ext2fs_instance));
  retval = sos_kmutex_init(& fs->lock, "ext2fs", SOS_KWQ_ORDER_FIFO);
  if (SOS_OK != retval)
    {
      sos_kfree((sos_vaddr_t) sb);
      sos_hash_dispose(hash);
      sos_kfree((sos_vaddr_t) fs);
      return retval;
    }
  fs->super.custom_data          = fs;
  fs->super.fs_type              = this;
  fs->super.allocate_new_node    = ext2fs_allocate_new_node;
  fs->super.fetch_node_from_disk = ext2fs_fetch_node_from_disk;
  fs->super.nodecache            = hash;
  fs->super.device               = device;
  fs->blocks_count     = sb->s_blocks_count;
  fs->blocks_per_group = sb->s_blocks_per_group;
  fs->inodes_per_group = sb->s_inodes_per_group;
  fs->block_size       = (1024 << sb->s_log_block_size);
  fs->device           = device->block_device;

  retval = ext2fs_fetch_node_from_disk(& fs->super,
				       EXT2_ROOT_INODE,
				       & fsnode_root);
  if (SOS_OK != retval)
    {
      sos_kfree((sos_vaddr_t) sb);
      sos_hash_dispose(hash);
      sos_kmutex_dispose(& fs->lock);
      sos_kfree((sos_vaddr_t) fs);
      return retval;
    }

  sos_bochs_printf ("root_fs_node=%p\n", fsnode_root);

  retval = sos_fs_register_fs_instance(& fs->super, fsnode_root);
  sos_fs_unref_fsnode(fsnode_root);
  if (SOS_OK != retval)
    {
      sos_kfree((sos_vaddr_t) sb);
      sos_hash_dispose(hash);
      sos_kmutex_dispose(& fs->lock);
      sos_kfree((sos_vaddr_t) fs);
      return retval;
    }

  *mounted_fs = & fs->super;
  return SOS_OK;
}


static sos_ret_t ext2fs_umount(struct sos_fs_manager_type * this,
			       struct sos_fs_manager_instance * mounted_fs)
{
  struct ext2fs_instance * ext2fs = (struct ext2fs_instance*)mounted_fs->custom_data;

  sos_hash_dispose(ext2fs->super.nodecache);
  sos_fs_unregister_fs_instance(& ext2fs->super);
  sos_kmutex_dispose(& ext2fs->lock);
  sos_kfree((sos_vaddr_t)ext2fs);
  return SOS_OK;
}

sos_ret_t sos_fs_ext2fs_subsystem_setup()
{
  strzcpy(ext2fs_type.name, "ext2fs", SOS_FS_MANAGER_NAME_MAXLEN);
  ext2fs_type.mount  = ext2fs_mount;
  ext2fs_type.umount = ext2fs_umount;

  return sos_fs_register_fs_type(& ext2fs_type);
}

/* ********************************************************
 * File mapping stuff
 */
inline static struct ext2fs_node *
get_ext2fsnode_of_vr(struct sos_umem_vmm_vr * vr)
{
  struct sos_umem_vmm_mapped_resource *mr
    = sos_umem_vmm_get_mapped_resource_of_vr(vr);

  return (struct ext2fs_node *)mr->custom_data;
}


static void ext2fs_map_ref(struct sos_umem_vmm_vr * vr)
{
  sos_bochs_printf("ext2fs_map_ref[%p]\n", vr);
  struct ext2fs_node * ext2fsnode = get_ext2fsnode_of_vr(vr);
  sos_fs_ref_fsnode(& ext2fsnode->super);
}


static void ext2fs_map_unref(struct sos_umem_vmm_vr * vr)
{
  sos_bochs_printf("ext2fs_map_unref[%p]\n", vr);
  struct ext2fs_node * ext2fsnode = get_ext2fsnode_of_vr(vr);
  _sos_fs_unref_fsnode(& ext2fsnode->super);
}

static sos_ret_t ext2fs_sync_dirty_page(sos_luoffset_t offset,
					sos_vaddr_t dirty_page,
					void * custom_data)
{
  sos_ret_t retval;
  struct ext2fs_node *ext2fsnode =
    (struct ext2fs_node *) custom_data;
  sos_size_t len = SOS_PAGE_SIZE;

  SOS_GENADDR_DECL(gaddr, FALSE, dirty_page);

  sos_bochs_printf("Flushing dirty page %u -> %llu\n",
		   dirty_page, offset);

  retval = ext2fs_contents_write ((struct ext2fs_instance *) ext2fsnode->super.fs,
				  ext2fsnode, offset, gaddr, & len, TRUE);
  if (SOS_OK != retval)
    return retval;

  if (SOS_PAGE_SIZE != len)
    return -SOS_EIO;

  sos_bochs_printf("End of flushing dirty page %u -> %llu\n",
		   dirty_page, offset);

  return SOS_OK;
}

static sos_ret_t ext2fs_map_page_in(struct sos_umem_vmm_vr * vr,
				    sos_uaddr_t uaddr,
				    sos_bool_t write_access)
{
  struct ext2fs_node * ext2fsnode;
  struct ext2fs_instance * fs;
  sos_uaddr_t    vr_start     = sos_umem_vmm_get_start_of_vr(vr);
  sos_luoffset_t start_offset = sos_umem_vmm_get_offset_in_resource(vr);
  struct sos_fs_pagecache_entry * pagecache_entry = NULL;
  sos_luoffset_t offset;
  sos_vaddr_t kernel_page;
  sos_bool_t read_from_disk = TRUE;
  sos_ret_t retval;

  ext2fsnode = get_ext2fsnode_of_vr(vr);
  fs         = (struct ext2fs_instance*) ext2fsnode->super.fs->custom_data;

  offset = SOS_PAGE_ALIGN_INF(uaddr) - vr_start;
  offset += start_offset;

  sos_bochs_printf ("ext2fs_map_page_in[vr=%p,uaddr=0x%x,%s]\n", vr, uaddr,
		    write_access ? "write" : "read");

  /* The region is not allowed to be resized */
  if (offset > SOS_PAGE_ALIGN_SUP(ext2fsnode->size))
    {
      sos_bochs_printf ("0x%x 0x%x %d", uaddr, vr_start, ext2fsnode->size);
      return -SOS_EFAULT;
    }

  if (sos_umem_vmm_get_flags_of_vr(vr) & SOS_VR_MAP_SHARED)
    {
      sos_bool_t newly_allocated;
      pagecache_entry = sos_fs_pagecache_ref_page(ext2fsnode->file.pc,
						  offset,
						  & kernel_page,
						  & newly_allocated);

      if (! pagecache_entry)
	return -SOS_EFAULT;

      if (! newly_allocated)
	read_from_disk = FALSE;
    }
  else
    {
      /* Allocate a temporary kernel page */
      kernel_page = sos_kmem_vmm_alloc(1, SOS_KMEM_VMM_MAP);
      if ((sos_vaddr_t)NULL == kernel_page)
	return -SOS_EFAULT;
    }

  retval = SOS_OK;

  if (read_from_disk)
    {
      SOS_GENADDR_DECL(gaddr, FALSE, kernel_page);
      sos_size_t rdlen = SOS_PAGE_SIZE;
      retval = ext2fs_contents_read (fs, ext2fsnode, offset, gaddr, &rdlen, FALSE);
      if (rdlen != SOS_PAGE_SIZE)
	{
	  sos_bochs_printf ("ext2fs_map_page_in(): could not read full page\n");
	  return -SOS_EIO;
	}
    }

  /* Make sure nobody mapped anything in the meantime */
  if ((sos_paddr_t)NULL != sos_paging_get_paddr(uaddr))
    retval = -SOS_EBUSY;

  if (SOS_OK == retval)
    {
      sos_paddr_t ppage_paddr = sos_paging_get_paddr(kernel_page);
      if (0 != ppage_paddr)
	retval = sos_paging_map(ppage_paddr,
				SOS_PAGE_ALIGN_INF(uaddr),
				TRUE,
				sos_umem_vmm_get_prot_of_vr(vr));
      else
	{
	  sos_bochs_printf ("No paddr found\n");
	  retval = -SOS_EFAULT;
	}
    }

  sos_bochs_printf("Pouet3\n");

  /* Shared mapping: unlock the entry and eventually remove it from
     the cache in case of failure of the preceding operations */
  if (NULL != pagecache_entry)
    {
      sos_fs_pagecache_unlock_page(ext2fsnode->file.pc,
				   pagecache_entry,
				   (SOS_OK != retval));
    }
  /* Private mappings: remove the remporary kernel page */
  else
    sos_kmem_vmm_free((sos_vaddr_t)kernel_page);

  return retval;
}


static struct sos_umem_vmm_vr_ops ext2fs_map_ops
  = (struct sos_umem_vmm_vr_ops){
    .ref     = ext2fs_map_ref,
    .unref   = ext2fs_map_unref,
    .page_in = ext2fs_map_page_in
  };

static sos_ret_t ext2fs_new_mapping(struct sos_umem_vmm_vr *vr)
{
  sos_bochs_printf ("ext2fs_new_mapping[%p]\n", vr);
  return sos_umem_vmm_set_ops_of_vr(vr, &ext2fs_map_ops);
}
// [[[]]]
