// [[[$article >= 11]]]
/* Copyright (C) 2005  Thomas Petazzoni

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.
*/

#include <sos/types.h>
#include <sos/klibc.h>

#include "netdev_generic.h"
#include "net_protos.h"

sos_ui16_t ipv4_checksum (const void *buf, unsigned int hdr_len) {
  unsigned long sum = 0;
  const sos_ui16_t *ip1;

  ip1 = buf;
  while ( hdr_len > 1 ) {
    sum += *ip1++;
    if ( sum & 0x80000000 )
      sum = ( sum & 0xFFFF ) + ( sum >> 16 );
    hdr_len -= 2;
  }

  while ( sum >> 16 )
    sum = ( sum & 0xFFFF ) + ( sum >> 16 );

  return ~sum;
}

struct sos_netdev_pkt * sos_ipv4_alloc_pkt (struct sos_netdev *dev,
					    sos_size_t len,
					    sos_bool_t is_for_xmit,
					    sos_ui32_t flags)
{
  struct sos_netdev_pkt *pkt;

  pkt = sos_eth_alloc_pkt (dev, len + sizeof (struct ip),
			   is_for_xmit, flags);
  if (NULL == pkt)
    return NULL;

  pkt->pdu_offset += sizeof (struct ip);
  pkt->pdu_len    -= sizeof (struct ip);

  sos_bochs_printf ("[IPV4] sos_ipv4_alloc_pkt(): %p (off=%d,len=%d)\n",
		    pkt->contents, pkt->pdu_offset, pkt->pdu_len);

  return pkt;
}

sos_ret_t sos_ipv4_recv_packet (struct sos_netdev * dev,
				struct sos_netdev_pkt * pkt)
{
  struct ip * ip_pkt;
  ip_pkt = (struct ip *) & pkt->contents[pkt->pdu_offset];

  /* Ignore packets that don't fit in a SINGLE netdev frame */
  if (ntohs(ip_pkt->length) > pkt->pdu_len)
    return -SOS_EINVAL;

  /* Ignore IPv6 frames */
  if (ip_pkt->version != 4)
    return -SOS_ENOSUP;

  /* Ignore malformed IPv4 headers */
  if (ip_pkt->header_length != 5)
    return -SOS_EINVAL;

  if (0 != memcmp(& ip_pkt->dest_addr, dev->netaddr, 4))
    /* Not for us => ignored, even if it is a broadcast ;) */
    return SOS_OK;

  pkt->pdu_offset += ip_pkt->header_length*4;
  pkt->pdu_len    -= ip_pkt->header_length*4;

  if (ip_pkt->protocol == IP_PROTO_ICMP)
    return sos_icmpv4_recv_packet (dev, ip_pkt, pkt);
  else if (ip_pkt->protocol == IP_PROTO_UDP)
    return sos_udp_recv_packet (dev, ip_pkt, pkt);
  else
    sos_bochs_printf ("Unsupported IP protocol: %d\n", ip_pkt->protocol);

  return SOS_OK;
}

sos_ret_t sos_ipv4_send_packet (struct sos_netdev * dev,
				struct sos_netdev_pkt * pkt,
				sos_ui32_t src_addr, sos_ui32_t dest_addr,
				sos_ui8_t protocol)
{
  struct ip * ip_pkt;

  pkt->pdu_offset -= sizeof(struct ip);
  pkt->pdu_len    += sizeof(struct ip);

  ip_pkt = (struct ip *) & pkt->contents[pkt->pdu_offset];

  sos_bochs_printf ("[IPV4] sos_ipv4_send_packet() %p %p (off=%d,len=%d)\n",
		    pkt->contents, ip_pkt, pkt->pdu_offset, pkt->pdu_len);

  memset (ip_pkt, 0, sizeof (struct ip));

  ip_pkt->version       = 4;
  ip_pkt->header_length = 5;
  ip_pkt->protocol      = protocol;
  ip_pkt->src_addr      = htonl(src_addr);
  ip_pkt->dest_addr     = htonl(dest_addr);
  ip_pkt->length        = htons(pkt->pdu_len);
  ip_pkt->checksum      = 0;
  ip_pkt->checksum      = ipv4_checksum (ip_pkt, sizeof(struct ip));

  return sos_arp_send_packet (dev, pkt, dest_addr, ETH_TYPE_IPv4);
}

// [[[]]]
