// [[[$article >= 11]]]
/* Copyright (C) 2005  Thomas Petazzoni

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.
*/

#include <sos/assert.h>
#include <sos/thread.h>
#include <sos/kmalloc.h>
#include <sos/list.h>
#include <hwcore/irq.h>

#include "netdev_generic.h"


struct sos_netdev_pkt * sos_netdev_alloc_pkt (struct sos_netdev * dev,
					      sos_size_t len,
					      sos_bool_t is_for_xmit,
					      sos_ui32_t alloc_flags)
{
  struct sos_netdev_pkt *pkt;

  /* Forbid sending too large packets */
  if (is_for_xmit && (len > dev->mtu))
    return NULL;

  pkt = (struct sos_netdev_pkt *)
    sos_kmalloc(sizeof (struct sos_netdev_pkt) + len,
		alloc_flags);
  if (NULL == pkt)
    return NULL;

  memset (pkt, 0, sizeof (struct sos_netdev_pkt));
  pkt->contents_size = len;
  pkt->pdu_offset    = 0;
  pkt->pdu_len       = len;

  sos_bochs_printf ("sos_netdev_alloc_pkt(): %p (off=%d,len=%d)\n",
		    pkt->contents, pkt->pdu_offset, pkt->pdu_len);

  return pkt;
}


static void sos_netdev_free_pkt (struct sos_netdev_pkt *pkt)
{
  sos_kfree ((sos_vaddr_t) pkt);
}


static void netdev_rx_thread (void *arg)
{
  struct sos_netdev * dev = (struct sos_netdev*)arg;

  while (1)
    {
      struct sos_netdev_pkt *cur;
      sos_ui32_t flags;

      sos_bochs_printf("RECV: Waiting for next netdev packet...\n");
      sos_ksema_down (& dev->recv.pkt_list_sem, NULL);
      sos_bochs_printf("RECV: Woken up by netdev packet !\n");

      /* Retrieve the first packet in the list */
      sos_disable_IRQs(flags);
      SOS_ASSERT_FATAL(! list_is_empty(dev->recv.pkt_list));
      cur = list_pop_head(dev->recv.pkt_list);
      sos_restore_IRQs(flags);

      sos_bochs_printf("RECV Handle pkt %p sz=%d\n", cur, cur->contents_size);

      if (dev->recv.netstack_recv
	  && (SOS_OK == dev->recv.netstack_recv(dev, cur)))
	{
	  sos_disable_IRQs(flags);
	  dev->nb_rx_pkts ++;
	  dev->nb_rx_bytes += cur->contents_size;
	  sos_restore_IRQs(flags);
	}

      sos_netdev_free_pkt(cur);
      sos_bochs_printf("RECV Handle end\n");
    }
}


sos_ret_t sos_netdev_received_pkt (struct sos_netdev * dev,
				   struct sos_netdev_pkt * pkt)
{
  sos_ui32_t flags;

  sos_disable_IRQs(flags);
  list_add_tail (dev->recv.pkt_list, pkt);
  sos_restore_IRQs(flags);

  /* Now Wakeup consummer thread */
  sos_ksema_up (& dev->recv.pkt_list_sem);

  return SOS_OK;
}


static void netdev_tx_thread (void *arg)
{
  struct sos_netdev * dev = (struct sos_netdev*)arg;

  while (1)
    {
      struct sos_netdev_pkt *cur;
      sos_ui32_t flags;

      sos_bochs_printf("SEND: Waiting for sending netdev packet...\n");
      sos_ksema_down (& dev->send.pkt_list_sem, NULL);
      sos_bochs_printf("SEND: Woken up by netdev send req !\n");

      /* Retrieve the first packet in the list */
      sos_disable_IRQs(flags);
      SOS_ASSERT_FATAL(! list_is_empty(dev->send.pkt_list));
      cur = list_pop_head(dev->send.pkt_list);
      sos_restore_IRQs(flags);

      sos_bochs_printf("SEND pkt %p sz=%d\n", cur, cur->contents_size);
      if (dev->send.device_send
	  && (SOS_OK == dev->send.device_send(dev, cur)))
	{
	  sos_disable_IRQs(flags);
	  dev->nb_tx_pkts ++;
	  dev->nb_tx_bytes += cur->contents_size;
	  sos_restore_IRQs(flags);
	}

      sos_netdev_free_pkt(cur);
      sos_bochs_printf("SEND Handle end\n");
    }
}


sos_ret_t sos_netdev_send_pkt (struct sos_netdev * dev,
			       struct sos_netdev_pkt *pkt)
{
  sos_ui32_t flags;

  sos_disable_IRQs(flags);
  list_add_tail (dev->send.pkt_list, pkt);
  sos_restore_IRQs(flags);

  /* Now Wakeup consummer thread */
  sos_ksema_up (& dev->send.pkt_list_sem);

  return SOS_OK;
}

static struct sos_netdev *default_netdev = NULL;
struct sos_netdev *sos_netdev_default (void)
{
  return default_netdev;
}

sos_ret_t sos_netdev_init (struct sos_netdev * dev,
			   const char * name,
			   sos_netdev_recv_handler_t * recv_upcall,
			   sos_netdev_send_handler_t * send_downcall)
{
  memset(dev, 0, sizeof(struct sos_netdev));
  strzcpy(dev->name, name, SOS_NETDEV_MAX_NAMELEN);

  list_init (dev->recv.pkt_list);
  sos_ksema_init (&  dev->recv.pkt_list_sem, "netdev_rx",
		  0, SOS_KWQ_ORDER_FIFO);
  dev->recv.netstack_recv = recv_upcall;

  list_init (dev->send.pkt_list);
  sos_ksema_init (&  dev->send.pkt_list_sem, "netdev_tx",
		  0, SOS_KWQ_ORDER_FIFO);
  dev->send.device_send   = send_downcall;

  if (NULL == sos_create_kernel_thread("[netdev_rx]",
				       netdev_rx_thread,
				       dev, SOS_SCHED_PRIO_RT_LOWEST))
    return -SOS_ENOMEM;

  if (NULL == sos_create_kernel_thread("[netdev_tx]",
				       netdev_tx_thread,
				       dev, SOS_SCHED_PRIO_RT_LOWEST))
    return -SOS_ENOMEM;

  default_netdev = dev;

  return SOS_OK;
}

// [[[]]]
