// [[[$article >= 11]]]
/* Copyright (C) 2005  Thomas Petazzoni

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.
*/

#ifndef __SOS_NETDEV_GENERIC_H__
#define __SOS_NETDEV_GENERIC_H__


#include <sos/types.h>
#include <sos/errno.h>
#include <sos/ksynch.h>


/**
 * @file netdev_generic.h
 *
 * Generic management of network device packets.
 *
 * A generic "network device" is defined and the context for packet
 * transmission/reception are defined. All this should be considered
 * as a skeleton for network device driver. A device driver:
 *  - handles the reception of packets and possibly assemble them in
 *    upper-layer packets. For example, an ethernet hardware device
 *    might receive ethernet packets but the device driver might only
 *    see the ethernet frame split into multiple chuncks. It must
 *    assemble them to form the actual ethernet frame before sending
 *    the latter to the upper layer, here the "ethernet" layer.
 *  - handles the transmission of upper-layer packets (eg ethernet frames).
 *
 * The netdev_generic API provides:
 *  - the definition of a generic network device (struct sos_netdev)
 *  - the infrastructure by which the device driver signals that a
 *    packet was received (sos_netdev_received_pkt)
 *  - the infrastructure by which the device driver is called by the
 *    upper layers to send data
 *
 * The latter 2 items take the form of 2 callbacks into the struct netdev:
 *  - sos_netdev::recv.handler called when the device driver received
 *    a frame. This function is never called direclty by the device
 *    driver: it is always called from inside a dedicated thread, which
 *    is awoken by sos_netdev_received_pkt(), and could possibly
 *    re-assemble the packets before calling the upper-layer reception
 *    function (eg ethernet)
 *  - sos_netdev::send.handler called when a frame is to be sent over
 *    the network. Called by the upper-layer (eg
 *    ethernet). Corresponds to a function implemented by the device
 *    driver.
 *
 * Typical packet reception chronogram:
 *
 * - IRQ -> device driver
 *   + device driver allocates a new packet structure of the correct size
 *   + device driver acknowledges the reception of the packet and the IRQ
 *   + device driver calls sos_netdev_received_pkt()
 *   + device driver exits from IRQ code
 * - sooner or later, the reception thread will be awoken
 *   + reception thread calls sos_netdev::recv.handler(dev, pkt) for
 *     each packet waiting and deletes the packet
 *   + reception thread sleeps until next sos_netdev_received_pkt()
 */


/* Forward declarations */
struct sos_netdev;
struct sos_netdev_pkt;

/** Generic handler called upon packet reception for reassembly and
    transmission to upper protocol layers */
typedef sos_ret_t (sos_netdev_recv_handler_t (struct sos_netdev *dev,
					      struct sos_netdev_pkt *pkt));

/** Generic handler called to send packets over the network */
typedef sos_ret_t (sos_netdev_send_handler_t (struct sos_netdev *dev,
					      struct sos_netdev_pkt *pkt));


struct sos_netdev
{
#define SOS_NETDEV_MAX_NAMELEN 16
  char name[SOS_NETDEV_MAX_NAMELEN];

#define SOS_NETDEV_MAX_HWADDR_LEN  32
  sos_ui8_t hwaddr[SOS_NETDEV_MAX_HWADDR_LEN];

  /** For simplicity reasons, we assume that each network device has
      only ONE network address */
#define SOS_NETDEV_MAX_NETADDR_LEN 32
  sos_ui8_t netaddr[SOS_NETDEV_MAX_NETADDR_LEN]; /**< NETWORK byte order */

  sos_ui32_t mtu;

  /** Data to manage the *reception* of device packets */
  struct
  {
    struct sos_netdev_pkt *pkt_list;
    struct sos_ksema pkt_list_sem;

    /* upcall to the upper layer protocol stack */
    sos_netdev_recv_handler_t *netstack_recv;

    /* Temporary data for frame reconstitution at reception (reserved
       to handler) */
    void * handler_data;
  } recv;


  /** Data to manage the *transmission* of network device packets */
  struct
  {
    struct sos_netdev_pkt *pkt_list;
    struct sos_ksema pkt_list_sem;

    /* downcall to the device driver */
    sos_netdev_recv_handler_t *device_send;

    /* Temporary data for frame reconstitution at reception (reserved
       to handler) */
    void * handler_data;
  } send;

  /* Some stats */
  sos_count_t nb_collisions;
  sos_count_t nb_errors;

  sos_count_t nb_rx_pkts;
  sos_count_t nb_tx_pkts;

  sos_lcount_t nb_rx_bytes;
  sos_lcount_t nb_tx_bytes;

  /** Data reserved for HW device driver */
  void * custom_data;
};


/**
 * Structure of a network device's packet:
 *   - When received: this corresponds to a an upper-layer PDU (aka
 *     ethernet, FDDI, ...) received, possibly after packet assembly by
 *     the device driver
 *   - When sent: this corresponds to an upper-layer PDU (ethernet,
 *     FDDI, ...) to be sent by the device
 */
struct sos_netdev_pkt {
  sos_size_t contents_size;
  struct sos_netdev_pkt *prev, *next;

  /* Actual data for current layer: this represents a PART of the contents ! */
  sos_uoffset_t pdu_offset;
  sos_size_t    pdu_len;

  /* This field MUST come last: it piggybacks the contents of the
     packet */
  char contents[0];
};


sos_ret_t sos_netdev_init(struct sos_netdev * dev,
			  const char *name,
			  sos_netdev_recv_handler_t * recv_upcall,
			  sos_netdev_send_handler_t * send_downcall);

struct sos_netdev *sos_netdev_default (void);

struct sos_netdev_pkt * sos_netdev_alloc_pkt (struct sos_netdev * dev,
					      sos_size_t len,
					      sos_bool_t is_for_xmit,
					      sos_ui32_t alloc_flags);

/** Called by the network device driver */
sos_ret_t sos_netdev_received_pkt (struct sos_netdev *dev,
				   struct sos_netdev_pkt *pkt);

/** Called by the upper layers */
sos_ret_t sos_netdev_send_pkt (struct sos_netdev *dev,
			       struct sos_netdev_pkt *pkt);


/* TODO: A deplacer */
sos_ret_t sos_arp_recv_packet (struct sos_netdev * dev,
			       struct sos_netdev_pkt * pkt);

sos_ret_t sos_eth_recv_packet (struct sos_netdev * dev,
			       struct sos_netdev_pkt *pkt);

sos_ret_t sos_ipv4_recv_packet (struct sos_netdev * dev,
				struct sos_netdev_pkt * pkt);

struct sos_netdev_pkt *sos_eth_alloc_pkt (struct sos_netdev * dev,
					  sos_size_t len,
					  sos_bool_t is_for_xmit,
					  sos_ui32_t kmalloc_flags);

sos_ret_t sos_eth_send_packet (struct sos_netdev *dev,
			       struct sos_netdev_pkt *pkt,
			       const sos_ui8_t *dstmac,
			       sos_ui16_t type);


#define htons(n) ( (((n) & 0xFF00) >> 8) | (((n) & 0x00FF) << 8) )
#define htonl(n) ( (((n) & 0xFF000000) >> 24) | (((n) & 0x00FF0000) >> 8) \
                | (((n) & 0x0000FF00) << 8) | (((n) & 0x000000FF) << 24))
#define ntohs(n) ( (((n) & 0xFF00) >> 8) | (((n) & 0x00FF) << 8) )
#define ntohl(n) ( (((n) & 0xFF000000) >> 24) | (((n) & 0x00FF0000) >> 8) \
                | (((n) & 0x0000FF00) << 8) | (((n) & 0x000000FF) << 24) )

#endif /* __SOS_NETDEV_GENERIC_H__ */
// [[[]]]
