// [[[$article >= 11]]]
/* Copyright (C) 2005  Thomas Petazzoni, David Decotigny

   Macro constants are:
   Copyright (c) 2003-2004 Fabrice Bellard
       Written 1993,1994,1999,2000 by Donald Becker.
       Copyright 1994,1999-2000 Donald Becker
       Copyright 1993 United States Government as represented by the
       Director, National Security Agency.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA. 
 */

#include <sos/macros.h>
#include <sos/list.h>
#include <sos/types.h>
#include <sos/kmalloc.h>
#include <sos/assert.h>
#include <sos/klibc.h>
#include <hwcore/ioports.h>
#include <hwcore/irq.h>
#include <drivers/pci.h>

#include "netdev_generic.h"

/**
 * @file ne2k.c
 *
 * PIO driver for NE200 compatible ISA cards
 *
 * This file is based on the NS 8390 specs
 * (http://www.national.com/search/search_wrap.cgi?textfield=8390), on
 * the RTL 8019(AS) register tables
 * (http://www.realtek.com.tw/downloads/downloads1-3.aspx?spec=True&compamodel=RTL8019),
 * on Donald Becker's 8390.c Linux driver, on Donald Becker's
 * diagnostic tool (ftp://ftp.scyld.com/pub/diag/ne2k-diag.c) and on
 * the qemu/bochs NE2000 emulation code
 */

/* Registers common to all pages */
#define ENX_CMD   0x00    /* The command register (for all pages) */

/* Registers for page 0 */
#define EN0_CLDALO  0x01    /* Low byte of current local dma addr  RD */
#define EN0_STARTPG 0x01    /* Starting page of ring bfr WR */
#define EN0_CLDAHI  0x02    /* High byte of current local dma addr  RD */
#define EN0_STOPPG  0x02    /* Ending page +1 of ring bfr WR */
#define EN0_BOUNDARY    0x03    /* Boundary page of ring bfr RD WR */
#define EN0_TPSR    0x04    /* Transmit starting page WR */
#define EN0_TCNTLO  0x05    /* Low  byte of tx byte count WR */
#define EN0_FIFO    0x06    /* FIFO RD */
#define EN0_TCNTHI  0x06    /* High byte of tx byte count WR */
#define EN0_ISR     0x07    /* Interrupt status reg RD WR */
#define EN0_CRDALO  0x08    /* low byte of current remote dma address RD */
#define EN0_RSARLO  0x08    /* Remote start address reg 0 */
#define EN0_CRDAHI  0x09    /* high byte, current remote dma address RD */
#define EN0_RSARHI  0x09    /* Remote start address reg 1 */
#define EN0_RCNTLO  0x0a    /* Remote byte count reg WR */
#define EN0_RCNTHI  0x0b    /* Remote byte count reg WR */
#define EN0_RXCR    0x0c    /* RX configuration reg WR */
#define EN0_TXCR    0x0d    /* TX configuration reg WR */
#define EN0_DCR     0x0e    /* Data configuration reg WR */
#define EN0_IMR     0x0f
#define EN0_DATAPORT 0x10

/* A little hack here: offset >= 0x10 corresponds to registers in page
   1. To access them, the CR is configured for page 1 access, the
   register offset-0x10 is accessed, and finally the CR is configured
   for access to page 0 again */
#define EN1_BASE_PHYS   0x11 /* Register 1 in page 1 == PAR0 */
#define EN1_CURPAG      0x17 /* Register 7 in page 1 == CURR */
#define EN1_BASE_MAR    0x18 /* Register 8 in page 1 == MAR0 */

/*  Bitmap for command register.  */
#define E8390_STOP  0x01    /* Stop and reset the chip */
#define E8390_START 0x02    /* Start the chip, clear reset */
#define E8390_TRANS 0x04    /* Transmit a frame */

#define E8390_RREAD  0x08    /* Remote read */
#define E8390_RWRITE 0x10    /* Remote write  */
#define E8390_NODMA  0x20    /* Abort remote DMA */

#define E8390_PAGE0 0x00    /* Select page chip registers */
#define E8390_PAGE1 0x40    /* using the two high-order bits */
#define E8390_PAGE2 0x80    /* Page 3 is invalid. */

/* Bits in EN0_ISR - Interrupt status register */
#define ENISR_RX    0x01    /* Receiver, no error */
#define ENISR_TX    0x02    /* Transmitter, no error */
#define ENISR_RX_ERR    0x04    /* Receiver, with error */
#define ENISR_TX_ERR    0x08    /* Transmitter, with error */
#define ENISR_OVER  0x10    /* Receiver overwrote the ring */
#define ENISR_COUNTERS  0x20    /* Counters need emptying */
#define ENISR_RDC   0x40    /* remote dma complete */
#define ENISR_RESET 0x80    /* Reset completed */
#define ENISR_ALL   0x3f    /* Interrupts we will enable */

/* Bitmap for EN0_RXCR */
#define E8390_RXCONFIG          0x4     /* EN0_RXCR: broadcasts, no multicast,errors */
#define E8390_RXOFF             0x20    /* EN0_RXCR: Accept no packets */

/* Bitmap for EN0_TXCR */
#define E8390_TXCONFIG          0x00    /* EN0_TXCR: Normal transmit mode */
#define E8390_TXOFF             0x02    /* EN0_TXCR: Transmitter off */

/* Configuration of the recv/transmit ring buffer pages */
#define TX_BUF_START 0x40
#define TX_PAGES     12
#define TX_BUF_END   (TX_BUF_START + TX_PAGES)

#define RX_BUF_START (TX_BUF_END)
#define RX_PAGES     12
#define RX_BUF_END   (RX_BUF_START + RX_PAGES)


/** Definition of an ISA NE2000 device */
struct ne2k_netdev
{
  struct sos_netdev generic;
  sos_ui16_t base_ioaddr;
  sos_ui8_t  irq_line;

  int drv_rx_curpage;

  struct sos_ksema tx_ack_sema;
  sos_bool_t tx_success;

  struct ne2k_netdev *prev, *next;
};


/** All NE2000 ISA devices are chained in order to easily locate the
    origin the IRQs */
static struct ne2k_netdev * ne2k_isa_device_list = NULL;


/**
 * An NE2000 packet located on (at most) 1 page begins with this
 * header. See "Storage format for receive sent packets" in NS
 *  specs
 */
struct ne2k_pkt_hdr
{
  sos_ui8_t  status; /* status */
  sos_ui8_t  next;   /* pointer to next packet. */
  sos_ui16_t count;  /* header + packet length in bytes */
} __attribute__((packed));


static void ne2k_outb (struct ne2k_netdev * dev,
		       sos_ui8_t value, sos_ui16_t ne2k_register)
{
  SOS_ASSERT_FATAL (ne2k_register <= EN1_BASE_MAR);

  if (ne2k_register <= EN0_DATAPORT)
    {
      outb (value, dev->base_ioaddr + ne2k_register);
    }

  else if (ne2k_register <= EN1_BASE_MAR)
    {
      /* For data access: switch to page 1, access the data, and
	 revert back to page 0 (always) */
      outb (E8390_NODMA|E8390_PAGE1, dev->base_ioaddr + ENX_CMD);
      outb (value, dev->base_ioaddr + ne2k_register - 0x10);
      outb (E8390_NODMA|E8390_PAGE0, dev->base_ioaddr + ENX_CMD);
    }
}


static sos_ui8_t ne2k_inb (struct ne2k_netdev * dev,
			   sos_ui16_t ne2k_register)
{
  sos_ui8_t val;

  SOS_ASSERT_FATAL (ne2k_register <= EN1_BASE_MAR);

  if (ne2k_register <= EN0_DATAPORT)
    {
      val = inb (dev->base_ioaddr + ne2k_register);
    }

  else if (ne2k_register <= EN1_BASE_MAR)
    {
      /* For data access: switch to page 1, access the data, and
	 revert back to page 0 (always) */
      outb (E8390_NODMA|E8390_PAGE1, dev->base_ioaddr + ENX_CMD);
      val = inb (dev->base_ioaddr + ne2k_register - 0x10);
      outb (E8390_NODMA|E8390_PAGE0, dev->base_ioaddr + ENX_CMD);
    }

  else
    val = 0xff;

  return val;
}


static void ne2k_read_memory (struct ne2k_netdev * dev,
			      sos_ui16_t pos, sos_ui8_t *dst, sos_size_t len)
{
  sos_ui16_t  xferstart;
  sos_size_t xfersize;
  sos_ui16_t cur;

  if (len <= 0)
    return;

  /* Align transfer start & size on a word-boundary */
  xferstart  = SOS_ALIGN_INF(pos, 2);
  xfersize   = SOS_ALIGN_SUP(pos + len - xferstart, 2);

  /* Configure NIC local DMA */
  ne2k_outb (dev, xfersize & 0xff , EN0_RCNTLO);
  ne2k_outb (dev, (xfersize >> 8) & 0xff, EN0_RCNTHI);
  ne2k_outb (dev, xferstart & 0xff, EN0_RSARLO);
  ne2k_outb (dev, (xferstart >> 8) & 0xff, EN0_RSARHI);

  /* Send start command */
  ne2k_outb (dev, (E8390_RREAD | E8390_START), ENX_CMD);

  /* Read the values transferred */
  for (cur = xferstart ; xfersize >= 1 ; xfersize -= 2, cur += 2)
    {
      sos_ui16_t v = inw(dev->base_ioaddr + EN0_DATAPORT);
      if (cur >= pos)
	dst[cur-pos] = v & 0xff;
      if (cur-pos+1 < len)
	dst[cur-pos + 1] = (v >> 8);
    }
}


static void ne2k_write_memory (struct ne2k_netdev * dev,
			       sos_ui16_t pos,
			       sos_ui8_t *src, sos_size_t src_len,
			       sos_size_t write_len)
{
  sos_ui16_t  xferstart;
  sos_size_t xfersize;
  sos_ui16_t cur;

  if (write_len <= 0)
    return;

  /* Align transfer start & size on a word-boundary */
  xferstart  = SOS_ALIGN_INF(pos, 2);
  xfersize   = SOS_ALIGN_SUP(pos + write_len - xferstart, 2);

  /* Configure NIC local DMA */
  ne2k_outb (dev, xfersize & 0xff , EN0_RCNTLO);
  ne2k_outb (dev, (xfersize >> 8) & 0xff, EN0_RCNTHI);
  ne2k_outb (dev, xferstart & 0xff, EN0_RSARLO);
  ne2k_outb (dev, (xferstart >> 8) & 0xff, EN0_RSARHI);

  /* Send start command */
  ne2k_outb (dev, (E8390_RWRITE | E8390_START), ENX_CMD);

  /* Write the values transferred */
  for (cur = xferstart ; xfersize >= 1 ; xfersize -= 2, cur += 2)
    {
      sos_ui16_t v = 0;

      if (cur-pos+1 < src_len)
	v |= src[cur-pos + 1];

      v <<= 8;
      if ((cur >= pos) && (cur-pos < src_len))
	v |= src[cur-pos];

      outw(v, dev->base_ioaddr + EN0_DATAPORT);
    }
}


static void ne2k_irq_handler_one_device (struct ne2k_netdev * dev,
					 sos_ui8_t ne2k_intr_status)
{
  sos_bochs_printf("[NE2K] dev %p IRQ ! (0x%x)\n", dev, ne2k_intr_status);

  if (ne2k_intr_status & (ENISR_TX | ENISR_TX_ERR))
    {
      dev->tx_success = !(ne2k_intr_status & ENISR_TX_ERR);
      sos_bochs_printf("[NE2K] ISR TX ACK %d\n", dev->tx_success);
      sos_ksema_up(& dev->tx_ack_sema);
      ne2k_outb (dev, ENISR_TX | ENISR_TX_ERR, EN0_ISR);
    }

  if (ne2k_intr_status & ENISR_RX)
    {
      struct ne2k_pkt_hdr hdr;
      sos_size_t pkt_len;

      /* Eat as many packets as available on the NIC */
      while (1)
	{
	  unsigned int nic_nextpage;
	  struct sos_netdev_pkt *pkt;

	  /* Acknowledge the reception of the packet */
	  ne2k_outb (dev, ENISR_RX | ENISR_RX_ERR, EN0_ISR);

	  /* Remember the position of the NIC in the ring buffer for
	     next received packet */
	  nic_nextpage = ne2k_inb (dev, EN1_CURPAG);

	  sos_bochs_printf ("[NE2K] Ack RX DRV CUR=0x%x, NIC NXT=0x%x\n",
			    dev->drv_rx_curpage, nic_nextpage);

	  /* Stop when there are no received packets left */
	  if (dev->drv_rx_curpage == nic_nextpage)
	    break;

	  ne2k_read_memory(dev, dev->drv_rx_curpage << 8, (sos_ui8_t*) & hdr,
			   sizeof(struct ne2k_pkt_hdr));
	  pkt_len = hdr.count - sizeof(struct ne2k_pkt_hdr);

	  sos_bochs_printf("[NE2K] Received packet, size=%d, cur=0x%x next=0x%x, status=0x%x\n",
			   pkt_len, dev->drv_rx_curpage, hdr.next, hdr.status);

	  pkt = sos_netdev_alloc_pkt(& dev->generic,
				     pkt_len,
				     FALSE, SOS_KMALLOC_ATOMIC);

	  ne2k_read_memory(dev, (dev->drv_rx_curpage << 8)
			   + sizeof(struct ne2k_pkt_hdr),
			   (sos_ui8_t*) pkt->contents, pkt_len);

	  sos_netdev_received_pkt(& dev->generic, pkt);

	  //	  ne2k_outb (dev, hdr.next-1, EN0_BOUNDARY);

	  ne2k_outb (dev, dev->drv_rx_curpage, EN0_BOUNDARY);
	  dev->drv_rx_curpage = hdr.next;
	}

    }

  if (ne2k_intr_status & ENISR_RDC)
    {
      ne2k_outb (dev, ENISR_RDC, EN0_ISR);
      sos_bochs_printf("[NE2K] activate RDC again\n");
    }

  ne2k_outb (dev, E8390_NODMA|E8390_PAGE0|E8390_START, ENX_CMD);
  sos_bochs_printf("[NE2K] END OF IRQ\n");
}


static void ne2k_isa_irq_handler (int irq_level)
{
  int nb;
  struct ne2k_netdev * dev;

  list_foreach(ne2k_isa_device_list, dev, nb)
    {
      sos_ui8_t intr_status;

      if (dev->irq_line != irq_level)
	continue;

      intr_status = ne2k_inb (dev, EN0_ISR);
      if (intr_status != 0)
	ne2k_irq_handler_one_device(dev, intr_status);
    }
}


static sos_ret_t ne2k_send_packet (struct sos_netdev * generic_dev,
				   struct sos_netdev_pkt * pkt)
{
  sos_ret_t retval;
  struct ne2k_netdev * dev = (struct ne2k_netdev*) generic_dev->custom_data;
  struct sos_time timeout;
  sos_ui32_t flags;

  /* NS says to pad to zero up to at least 46 bytes */
  int write_len = pkt->contents_size;
  if (write_len < 46)
    write_len = 46;

  sos_bochs_printf("[NE2K] SENDING %d bytes...\n", write_len);

  sos_disable_IRQs(flags);

  /* Store the packet into the NIC memory */
  ne2k_write_memory(dev, TX_BUF_START << 8,
		    (sos_ui8_t*) pkt->contents, pkt->contents_size, write_len);

  /* Configure NIC to send it */
  ne2k_outb (dev, (write_len & 0xff), EN0_TCNTLO);
  ne2k_outb (dev, (write_len >> 8), EN0_TCNTHI);
  ne2k_outb (dev, TX_BUF_START, EN0_RSARHI);
  ne2k_outb (dev, 0, EN0_RSARLO);

  /* Now send it */
  dev->tx_success = FALSE;
  ne2k_outb (dev, E8390_NODMA|E8390_TRANS|E8390_START, ENX_CMD);
  sos_restore_IRQs(flags);

  sos_bochs_printf("[NE2K] SENT ?\n");
  /* Waiting for TX acknowledge */
  timeout.sec     = 0;
  timeout.nanosec = 15*1000*1000; /* 15ms */

  retval = sos_ksema_down(& dev->tx_ack_sema, & timeout);
  if (SOS_OK != retval)
    sos_bochs_printf("[NE2K] timeout in send\n");

  sos_bochs_printf("[NE2K] SEND COMPLETED %x %d\n", retval, dev->tx_success);
  if ((SOS_OK == retval) && (! dev->tx_success))
    retval = -SOS_EBUSY;

  return retval;
}

struct sos_netdev * sos_ne2k_isa_init(const char * name,
				      sos_ui16_t base_ioaddr,
				      sos_ui8_t  irq_line,
				      sos_ui32_t myip /* Host byte order */)
{
  sos_ret_t retval;
  struct ne2k_netdev * dev;
  unsigned char SA_prom[32];
  int i;

  /* Allocate a new device description */
  dev = (struct ne2k_netdev*)
    sos_kmalloc(sizeof(struct ne2k_netdev), 0);
  if (! dev)
    return NULL;

  /* Initialize it */
  retval = sos_netdev_init(& dev->generic, name,
			   sos_eth_recv_packet,
			   ne2k_send_packet);
  if (SOS_OK != retval)
    {
      sos_kfree((sos_vaddr_t)dev);
      return NULL;
    }

  /* Setup the reserved fields */
  dev->generic.custom_data = (void*) dev;
  sos_ksema_init(& dev->tx_ack_sema, "ne2k_tx_ack", 0, SOS_KWQ_ORDER_FIFO);
  dev->generic.mtu = 1500; /* Ethernet "standard" MTU */
  dev->irq_line    = irq_line;
  dev->base_ioaddr = base_ioaddr;
  dev->drv_rx_curpage     = 0;

  myip = htonl(myip);  memcpy(& dev->generic.netaddr, & myip, 4);
  list_add_tail(ne2k_isa_device_list, dev);

  /* Reinitialize the NIC (hard reset + abort DMA + select Page 0) */
  ne2k_outb (dev, (E8390_NODMA | E8390_PAGE0 | E8390_STOP), ENX_CMD);

  /* Set 16b-wide accesses + no loopback */
  ne2k_outb (dev, 0x49, EN0_DCR);

  /* Reset remote byte count registers */
  ne2k_outb (dev, 0, EN0_RCNTLO);
  ne2k_outb (dev, 0, EN0_RCNTHI);

  /* Reset Receive register */
  ne2k_outb (dev, 0, EN0_RXCR);

  /* Disable IRQ handling during initialization */
  ne2k_outb (dev, 0, EN0_IMR);    /* Forbid all IRQs */
  ne2k_outb (dev, 0xff, EN0_ISR); /* Clear any pending IRQ */

  /* Disable RX & TX */
  ne2k_outb (dev, E8390_RXOFF, EN0_RXCR);
  ne2k_outb (dev, E8390_TXOFF, EN0_TXCR);

  /* Read the first 32 bytes of the PROM */
  ne2k_read_memory(dev, 0, SA_prom, sizeof(SA_prom));

  /* Read the MAC address of the network card, which is located at odd
     addresses at the beginning of the card memory */
  for (i = 0; i < 6; i++)
    dev->generic.hwaddr[i] = SA_prom[i*2];

  sos_bochs_printf("[NE2K] macaddr=%x:%x:%x:%x:%x:%x\n",
		   dev->generic.hwaddr[0], dev->generic.hwaddr[1],
		   dev->generic.hwaddr[2], dev->generic.hwaddr[3],
		   dev->generic.hwaddr[4], dev->generic.hwaddr[5]);

  /*
   * Now really initialize the card
   */

  /* Reinitialize the NIC (hard reset + abort any DMA + select Page 0) */
  ne2k_outb (dev, (E8390_NODMA | E8390_PAGE0 | E8390_STOP), ENX_CMD);

  /* Set 16b-wide accesses + no loopback */
  ne2k_outb (dev, 0x49, EN0_DCR);

  /* Reset read count registers */
  ne2k_outb (dev, 0, EN0_RCNTLO);
  ne2k_outb (dev, 0, EN0_RCNTHI);

  /* Set start/stop pages for TX and RX */
  ne2k_outb (dev, TX_BUF_START,  EN0_TPSR);
  ne2k_outb (dev, RX_BUF_START,  EN0_STARTPG);
  ne2k_outb (dev, RX_BUF_END-1,  EN0_BOUNDARY);
  ne2k_outb (dev, RX_BUF_END,    EN0_STOPPG);
  dev->drv_rx_curpage = RX_BUF_START;

  /* Configure transmission/reception */
  ne2k_outb (dev, E8390_RXCONFIG, EN0_RXCR);
  ne2k_outb (dev, E8390_TXCONFIG, EN0_TXCR);

  //  ne2k_outb (dev, ENISR_RDC|0x1|0x4, EN0_ISR);

  /* Copy MAC address in ROM to 8390 PA registers */
  for (i = 0 ; i < 6 ; i++)
    {
      unsigned char c;
      ne2k_outb (dev, dev->generic.hwaddr[i], EN1_BASE_PHYS + i);

      /* Make sure mac is correctly written */
      c = ne2k_inb (dev, EN1_BASE_PHYS + i);
      SOS_ASSERT_FATAL(c == dev->generic.hwaddr[i]);
    }

  /* Set current page */
  ne2k_outb (dev, dev->drv_rx_curpage, EN1_CURPAG);

  /* Setup the IRQ routine for reception signal and end of
     transmission signal */
  sos_irq_set_routine (irq_line, ne2k_isa_irq_handler);

  /* Now enable interrupts */
  ne2k_outb (dev, 0xFF, EN0_ISR); /* Reset pending IRQ */
  ne2k_outb (dev, 0x3F, EN0_IMR); /* Unmask all IRQ (except RDMA complete,
				which cause the ISR to reenter
				themselves...) ! */
  ne2k_outb (dev, E8390_NODMA|E8390_PAGE0|E8390_START, ENX_CMD); /* Go ! */

  return & dev->generic;
}

struct sos_netdev * sos_ne2k_pci_init(const char * name,
				      sos_ui32_t myip)
{
  struct sos_pci_device *card = NULL;

  /* PCI test */
  sos_bool_t sos_ne2k_pci_find (struct sos_pci_device *device, void *data)
    {
      card = device;
      return FALSE;
    }

  sos_pci_scan_devices (0x10ec, 0x8029, sos_ne2k_pci_find, TRUE, NULL);

  if (! card)
    return NULL;

  if (card->resource[0].type != SOS_PCI_RESOURCE_IO ||
      card->resource[0].base_address == 0 ||
      card->resource[0].length       == 0)
    return NULL;

  return sos_ne2k_isa_init (name, card->resource[0].base_address,
			    card->irq_line, myip);
}

// [[[]]]
