// [[[$enclosing && $article >= 11]]]

/* Copyright (C) 2005      David Decotigny

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.
*/

#include <sos/types.h>
#include <hwcore/ioports.h>
#include <drivers/bochs.h>
#include <drivers/pci.h>
#include <sos/kmalloc.h>
#include <sos/klibc.h>
#include <sos/list.h>

/*
  PCI specs chapter 6
  http://www.mega-tokyo.com/osfaq2/index.php/Where%20can%20I%20find%20programming%20info%20on%20PCI%3F
  http://my.execpc.com/~geezer/osd/pnp/pci.c
  http://cvs.sourceforge.net/viewcvs.py/clicker/c32-lxsdk/kernel/src/mods/io/pci/pcicore.c?view=markup
  http://www.acm.uiuc.edu/sigops/roll_your_own/7.c.0.html
*/

#define IOPORT_PCI_CFG_ADDR  0xcf8
#define IOPORT_PCI_CFG_DATA  0xcfc

union pci_cfg_addr
{
  struct pci_cfg_addr_bitfield
  {
    sos_ui32_t reg       :8;
    sos_ui32_t function  :3;
    sos_ui32_t device    :5;
    sos_ui32_t bus       :8;
    sos_ui32_t reserved  :7;
    sos_ui32_t enable    :1;
  } bitfield;
  sos_ui32_t ui32;
};

static struct sos_pci_device *pci_device_list = NULL;

sos_ui8_t pci_read_cfg8(int bus, int device, int function, int reg)
{
  union pci_cfg_addr addr;
  sos_ui8_t v;
  addr.bitfield = (struct pci_cfg_addr_bitfield) {
    .reg      = (reg & ~3),
    .function = function,
    .device   = device,
    .bus      = bus,
    .reserved = 0,
    .enable   = 1
  };

  outl(addr.ui32, IOPORT_PCI_CFG_ADDR);
  v = inb(IOPORT_PCI_CFG_DATA + (reg & 3));
  //  sos_bochs_printf("inb port %x => %x\n", addr.ui32, v);
  return v;
}


void pci_write_cfg8(int bus, int device, int function, int reg,
		    sos_ui8_t value)
{
  union pci_cfg_addr addr;
  addr.bitfield = (struct pci_cfg_addr_bitfield) {
    .reg      = (reg & ~3),
    .function = function,
    .device   = device,
    .bus      = bus,
    .reserved = 0,
    .enable   = 1
  };

  outl(addr.ui32, IOPORT_PCI_CFG_ADDR);
  outb(value, IOPORT_PCI_CFG_DATA + (reg & 3));
}


sos_ui32_t pci_read_cfg32(int bus, int device, int function, int reg)
{
  union pci_cfg_addr addr;
  sos_ui32_t v;
  addr.bitfield = (struct pci_cfg_addr_bitfield) {
    .reg      = reg,
    .function = function,
    .device   = device,
    .bus      = bus,
    .reserved = 0,
    .enable   = 1
  };

  outl(addr.ui32, IOPORT_PCI_CFG_ADDR);
  v = inl(IOPORT_PCI_CFG_DATA);
  //  sos_bochs_printf("inb port %x => %x\n", addr.ui32, v);
  return v;
}


void pci_write_cfg32(int bus, int device, int function, int reg,
		    sos_ui32_t value)
{
  union pci_cfg_addr addr;
  addr.bitfield = (struct pci_cfg_addr_bitfield) {
    .reg      = reg,
    .function = function,
    .device   = device,
    .bus      = bus,
    .reserved = 0,
    .enable   = 1
  };

  outl(addr.ui32, IOPORT_PCI_CFG_ADDR);
  outl(value, IOPORT_PCI_CFG_DATA);
}


void pci_scan(void)
{
  int bus;
  for (bus = 0 ; bus < 8 ; bus ++)
    {
      int dev;
      for (dev = 0 ; dev < 32 ; dev ++)
	{
	  sos_ui8_t header_type = pci_read_cfg8(bus, dev, 0, 0x0e);
	  sos_ui16_t vid
	    = pci_read_cfg8(bus, dev, 0, 0)
	    | (pci_read_cfg8(bus, dev, 0, 1) << 8);
	  sos_ui16_t devid
	    = pci_read_cfg8(bus, dev, 0, 2)
	    | (pci_read_cfg8(bus, dev, 0, 3) << 8);
	  sos_ui32_t class
	    = pci_read_cfg8(bus, dev, 0, 10)
	    | (pci_read_cfg8(bus, dev, 0, 11) << 8);

	  int nfuncs, func;

	  if ((vid == 0) || (vid == 0xffff))
	    continue;

	  sos_bochs_printf("PCI %d/%d %x:%x class=%lx header_type=%x\n",
			   bus, dev,
			   vid, devid,
			   class, header_type);

	  nfuncs = 8;
	  if (! (header_type & 0x80))
	    nfuncs = 1;

	  for (func = 0 ; func < nfuncs ; func ++)
	    {
	      sos_ui16_t vid
		= pci_read_cfg8(bus, dev, func, 0)
		| (pci_read_cfg8(bus, dev, func, 1) << 8);
	      sos_ui16_t devid
		= pci_read_cfg8(bus, dev, func, 2)
		| (pci_read_cfg8(bus, dev, func, 3) << 8);
	      sos_ui32_t class
		= pci_read_cfg8(bus, dev, func, 10)
		| (pci_read_cfg8(bus, dev, func, 11) << 8);
	      sos_ui16_t cmd
		= pci_read_cfg8(bus, dev, func, 4)
		| (pci_read_cfg8(bus, dev, func, 5) << 8);
	      sos_ui8_t irq_line
		= pci_read_cfg8(bus, dev, func, 0x3c);

	      int bar;

	      if ((vid == 0) || (vid == 0xffff))
		break;

	      sos_bochs_printf("    %x:%x func %d class=%lx header_type=%x cmd=%x, irq=%x\n",
			       vid, devid, func,
			       class, header_type, cmd, irq_line);
	      for (bar = 0 ; bar < 6 ; bar ++)
		{
		  sos_ui32_t baddr, size, barval
		    = pci_read_cfg32(bus, dev, func, 0x10 + (bar * 4));
		  pci_write_cfg32(bus, dev, func, 0x10 + (bar * 4),
				  0xffffffff);
		  size = pci_read_cfg32(bus, dev, func, 0x10 + (bar * 4));
		  pci_write_cfg32(bus, dev, func, 0x10 + (bar * 4),
				  barval);

		  if (barval == 0)
		    continue;

		  if (barval & 0x1)
		    {
		      /* I/O space */
		      baddr = barval & ~3;
		      size &= ~3;
		    }
		  else
		    {
		      /* mem */
		      baddr = barval & ~0xf;
		      size &= ~0xf;
		    }
		  size = ~size;

		  if (barval & 0x1)
		    size &= 0xffff;

		  sos_bochs_printf("      BAR%d: %s %lx-%lx\n",
				   bar, (barval & 0x1)?"io":"mem",
				   baddr, baddr + size);
		}
	    }
	}
    }
}

sos_ret_t sos_pci_scan_devices(sos_ui32_t filter_vendor_id,
			       sos_ui32_t filter_product_id,
			       sos_pci_scan_function_t scan_fct,
			       sos_bool_t only_unhandled_devices,
			       void *custom_data)
{
  struct sos_pci_device *dev;
  sos_bool_t ret;
  int n;

  list_foreach (pci_device_list, dev, n)
    {
      if (dev->vendor_id == filter_vendor_id   &&
	  dev->product_id == filter_product_id &&
	  ((only_unhandled_devices && ! dev->is_handled) ||
	   (! only_unhandled_devices)))
	{
	  ret = scan_fct (dev, custom_data);
	  if (! ret)
	    return SOS_OK;
	}
    }

  return SOS_OK;
}

sos_ret_t sos_pci_subsystem_setup(void)
{
  int bus;
  for (bus = 0 ; bus < 8 ; bus ++)
    {
      int dev;
      for (dev = 0 ; dev < 32 ; dev ++)
	{
	  struct sos_pci_device *device;

	  sos_ui8_t header_type = pci_read_cfg8(bus, dev, 0, 0x0e);
	  sos_ui16_t vid
	    = pci_read_cfg8(bus, dev, 0, 0)
	    | (pci_read_cfg8(bus, dev, 0, 1) << 8);
	  sos_ui16_t devid
	    = pci_read_cfg8(bus, dev, 0, 2)
	    | (pci_read_cfg8(bus, dev, 0, 3) << 8);
	  sos_ui32_t class
	    = pci_read_cfg8(bus, dev, 0, 10)
	    | (pci_read_cfg8(bus, dev, 0, 11) << 8);

	  int nfuncs, func;

	  if ((vid == 0) || (vid == 0xffff))
	    continue;

	  sos_bochs_printf("PCI %d/%d %x:%x class=%lx header_type=%x\n",
			   bus, dev,
			   vid, devid,
			   class, header_type);

	  nfuncs = 8;
	  if (! (header_type & 0x80))
	    nfuncs = 1;

	  for (func = 0 ; func < nfuncs ; func ++)
	    {
	      sos_ui16_t vid
		= pci_read_cfg8(bus, dev, func, 0)
		| (pci_read_cfg8(bus, dev, func, 1) << 8);
	      sos_ui16_t devid
		= pci_read_cfg8(bus, dev, func, 2)
		| (pci_read_cfg8(bus, dev, func, 3) << 8);
	      sos_ui32_t class
		= pci_read_cfg8(bus, dev, func, 10)
		| (pci_read_cfg8(bus, dev, func, 11) << 8);
	      sos_ui16_t cmd
		= pci_read_cfg8(bus, dev, func, 4)
		| (pci_read_cfg8(bus, dev, func, 5) << 8);
	      sos_ui8_t irq_line
		= pci_read_cfg8(bus, dev, func, 0x3c);

	      int bar;

	      if ((vid == 0) || (vid == 0xffff))
		break;

	      device = (struct sos_pci_device *)
		sos_kmalloc (sizeof(struct sos_pci_device), 0);
	      if (! device)
		return -SOS_ENOMEM;

	      memset (device, 0, sizeof (struct sos_pci_device));

	      device->bus_id     = bus;
	      device->device_id  = dev;
	      device->is_handled = FALSE;
	      device->vendor_id  = vid;
	      device->product_id = devid;
	      device->class_code = class;
	      device->irq_line   = irq_line;

	      sos_bochs_printf("    %x:%x func %d class=%lx header_type=%x cmd=%x, irq=%x\n",
			       vid, devid, func,
			       class, header_type, cmd, irq_line);
	      for (bar = 0 ; bar < 6 ; bar ++)
		{
		  sos_ui32_t baddr, size, barval
		    = pci_read_cfg32(bus, dev, func, 0x10 + (bar * 4));
		  pci_write_cfg32(bus, dev, func, 0x10 + (bar * 4),
				  0xffffffff);
		  size = pci_read_cfg32(bus, dev, func, 0x10 + (bar * 4));
		  pci_write_cfg32(bus, dev, func, 0x10 + (bar * 4),
				  barval);

		  if (barval == 0)
		    continue;

		  if (barval & 0x1)
		    {
		      /* I/O space */
		      baddr = barval & ~3;
		      size &= ~3;
		      device->resource[bar].type = SOS_PCI_RESOURCE_IO;
		    }
		  else
		    {
		      /* mem */
		      baddr = barval & ~0xf;
		      size &= ~0xf;
		      device->resource[bar].type = SOS_PCI_RESOURCE_MEM;
		    }
		  size = ~size;

		  if (barval & 0x1)
		    size &= 0xffff;

		  device->resource[bar].base_address = baddr;
		  device->resource[bar].length = size;

		  sos_bochs_printf("      BAR%d: %s %lx-%lx\n",
				   bar, (barval & 0x1)?"io":"mem",
				   baddr, baddr + size);
		}

	      list_add_tail (pci_device_list, device);
	    }
	}
    }

  return SOS_OK;
}

// [[[]]]
