// [[[ $master ]]]

// --------------------- Test parameters
// #define _CR "\r"
#define _CR "\n"

#ifdef KERNEL_SOS
# include <sos/klibc.h>
# include <drivers/bochs.h>
# include <drivers/x86_videomem.h>
# include "types.h"
# include "assert.h"
# include "list.h"
# include "kmem_slab.h"
# include "kmalloc.h"
# include <sos/physmem.h>
/* # define myprintf sos_bochs_printf */
# undef _CR
# define _CR
# define myprintf(a...) sos_x86_videomem_printf(10, 0, 7, a)

# define kmalloc(sz) sos_kmalloc((sz), SOS_KMALLOC_ATOMIC)
# define kfree(ptr)  sos_kfree((sos_vaddr_t)(ptr))
# define MALLOC_TEST test_art5
static void _yield(void)
{
// [[[$enclosing && $article >= 6.5]]]
  extern void sos_thread_yield(void);
  sos_thread_yield();
// [[[]]]
}
# define YIELD() _yield()

#else

# include <assert.h>
# include <stdlib.h>
# include <stdio.h>
# define SOS_ASSERT_FATAL assert
# define myprintf printf
typedef unsigned int sos_vaddr_t;
typedef unsigned int sos_size_t;
# define kmalloc malloc
# define kfree   free
# define MALLOC_TEST main
# define sos_physmem_get_state(str) ({ })
# define YIELD() ({ })
#endif

extern void sos_kmem_vmm_dump_vr(char const* str);

/*
 * kmalloc/kfree tests. To be run on a x86 ELF Posix host (typically: Linux).
 *
 * Allocates NB_ALLOC_MAX elements of random sizes (inside [4,
 * 64k]). Then frees a random number of these elements, chosen at
 * random, then re-allocate this same number of elements, with other
 * randomly chosen sizes, then frees, then allocate, then frees, then
 * ...
 *
 * Inbetween each step of each alloc/free campaign, the consistency of
 * the memory is checked (comparison with a unique magic for each
 * currently allocated element).
 */

// Maximum number of objects allocated at any point in time:
#define NB_ALLOC_MAX 50000
// Sizes of the objects
#define SZ_ALLOC_MIN 4
#define SZ_ALLOC_MAX 129000
// Number of alloc/free campaigns
#define NB_CAMPAIGNS 4
// Define this if you want to check every bytes in the allocated areas
#define CHECK_FULL_AREAS

static unsigned int current_alloc_size = 0, max_alloc_size = 0;

/* A simple way of issueing a random number in [0, max[. This is an
   incorrect way of doing it according to the "Numerical Recipes in C:
   The Art of Scientific Computing" (William H.  Press, Brian
   P. Flannery, Saul A. Teukolsky, William T. Vetterling; New York:
   Cambridge University Press, 1992 (2nd ed., p. 277)). But, well, we
   use it here because we don't have FPU support in kernel, and we'd
   rather not use 64bits arithmetic. */
int rnd_int1(int max)
{
  int ret = random() % max;
  SOS_ASSERT_FATAL((ret >= 0) && (ret < max));
  return ret;
}

int rnd_int2(int min, int max)
{
  int ret = min + (random() % (max - min + 1));
  SOS_ASSERT_FATAL((ret >= min) && (ret <= max));
  return ret;
}

/* Generate a randomized 0..nb_items-1 array */
void rnd_array(int *array, int nb_items)
{
  int i;

  for (i = 0 ; i < nb_items ; i++)
    array[i] = i;

  for (i = 0 ; i < nb_items ; i++)
    {
      /* Choose an array entry we want to swap with */
      int j = rnd_int1(nb_items);

      /* Actually swap them */
      int tmp = array[i];
      array[i] = array[j];
      array[j] = tmp;
    }
}


// --------------------- The tests
struct alloc_table_t {
  sos_vaddr_t base_addr;
  sos_size_t size;
} alloc_table[NB_ALLOC_MAX]; /* In bss => 0B in the kernel image file */

int free_array[NB_ALLOC_MAX];

int test_campaign(int free_len)
{
  int i;

  SOS_ASSERT_FATAL((free_len >=0) && (free_len <= NB_ALLOC_MAX));

  myprintf("# alloc/free campaign for %d elements..\n", free_len);
  rnd_array(free_array, NB_ALLOC_MAX);

  myprintf(_CR "kmalloc test...               ");
  for (i = 0 ; i < NB_ALLOC_MAX ; i++) {
    _yield();
    if (alloc_table[i].base_addr)
      continue; /* Already allocated */
    alloc_table[i].size      = rnd_int2(SZ_ALLOC_MIN, SZ_ALLOC_MAX);
    alloc_table[i].base_addr = (sos_vaddr_t)kmalloc(alloc_table[i].size);
    if (! alloc_table[i].base_addr)
      myprintf("out of mem for sz=%d (iterartion %d)\n",
	       alloc_table[i].size, i);
    else
      {
#ifdef CHECK_FULL_AREAS
	int j;
	for (j = 0 ; j*sizeof(int) < alloc_table[i].size ; j++)
	  {
	    // sos_bochs_printf("Write alloc[%d] integer %d\n", i, j);
	    ((int*)alloc_table[i].base_addr)[j] = i;
	  }
#else
	*((int*)alloc_table[i].base_addr) = i;
#endif
	current_alloc_size += alloc_table[i].size;
      }

    /* Tente de liberer un coup sur 3 */
    if (rnd_int1(3) == 0)
    {
      int index = free_array[rnd_int1(NB_ALLOC_MAX)];
      if (alloc_table[index].base_addr)
	{
	  if (*((int*)alloc_table[index].base_addr) != index)
	    {
	      myprintf("alloc[%d]=0x%x - Wrong magic\n", index,
		       alloc_table[index].base_addr);
	      return -1;
	    }
	  kfree((void*)alloc_table[index].base_addr);
	  current_alloc_size -= alloc_table[index].size;
	  alloc_table[index].size      = -1;
	  alloc_table[index].base_addr = 0;
	}
    }

  }

  if (current_alloc_size > max_alloc_size)
    max_alloc_size = current_alloc_size;

  myprintf(_CR "Check alloc...               ");
  for (i = 0 ; i < NB_ALLOC_MAX ; i++)
    {
      _yield();
      if (! alloc_table[i].base_addr)
	{
	  // myprintf("(alloc[%d] has not been allocated)\n", i);
	  continue;
	}
#ifdef CHECK_FULL_AREAS
      {
	int j;
	for (j = 0 ; j*sizeof(int) < alloc_table[i].size ; j++)
	  if (((int*)alloc_table[i].base_addr)[j] != i)
#else
	if (*((int*)alloc_table[i].base_addr) != i)
#endif
	  {
	    myprintf("alloc[%d]=0x%x - Wrong magic (is: %d, should be: %d)\n",
		     i, alloc_table[i].base_addr,
		     *((int*)alloc_table[i].base_addr), i);
	    return -1;
	  }
#ifdef CHECK_FULL_AREAS
      }
#endif
    }

  myprintf(_CR "Free...                 ");
  for (i = 0 ; i < free_len ; i++) {
    int index = free_array[i];
    _yield();
    if (! alloc_table[index].base_addr)
      continue;
    if (*((int*)alloc_table[index].base_addr) != index)
      {
	myprintf("alloc[%d]=0x%x - Wrong magic\n", index,
		 alloc_table[index].base_addr);
	return -1;
      }
    kfree((void*)alloc_table[index].base_addr);
    current_alloc_size -= alloc_table[index].size;
    alloc_table[index].size      = -1;
    alloc_table[index].base_addr = 0;
  }

  myprintf(_CR "Check alloc(2)...           ");
  for (i = 0 ; i < NB_ALLOC_MAX ; i++) {
    _yield();
    if (! alloc_table[i].base_addr) {
      //      myprintf("(alloc[%d] has not been allocated)\n", i);
      continue;
    }
#ifdef CHECK_FULL_AREAS
      {
	int j;
	for (j = 0 ; j*sizeof(int) < alloc_table[i].size ; j++)
	  if (((int*)alloc_table[i].base_addr)[j] != i)
#else
	if (*((int*)alloc_table[i].base_addr) != i)
#endif
	  {
	    myprintf("alloc[%d]=0x%x - Wrong magic (is: %d, should be: %d)\n",
		     i, alloc_table[i].base_addr,
		     *((int*)alloc_table[i].base_addr), i);
	    return -1;
	  }
#ifdef CHECK_FULL_AREAS
      }
#endif
  }
  
  // sos_kmem_vmm_dump_vr("Test campaign");

  return 0;
}


static void kmem_cache_test(void)
{
  struct toto { int idx; char unused[1234]; struct toto *prev, *next; } *lst = NULL, *elt;
  unsigned nb_alloc = 0, nb_free = 0;
  struct sos_kslab_cache *cache =
    sos_kmem_cache_create("Rien", sizeof(struct toto), 4, 12,
			  SOS_KSLAB_CREATE_MAP);
  SOS_ASSERT_FATAL(cache != NULL);

  while (1)
    {
      _yield();
      elt = (struct toto*) sos_kmem_cache_alloc(cache, 0);
      if (! elt)
	break;
      nb_alloc ++;
      elt->idx = nb_alloc;
      list_add_tail(lst, elt);

      if ((nb_alloc % 100) == 0)
	myprintf(_CR "Allocated %d objs", nb_alloc);

      SOS_ASSERT_FATAL(sos_kmem_cache_destroy(cache) != SOS_OK);
    }

/*   list_collapse(lst, elt) */
  while(1)
    {
      _yield();
      elt = (struct toto*)list_get_head(lst);
      if (! elt)
	break;

      SOS_ASSERT_FATAL(elt->idx == nb_free + 1);
      elt->idx = (unsigned)-1;
      list_delete(lst, elt);

      sos_kmem_cache_free((sos_vaddr_t)elt);

      nb_free ++;
      if ((nb_free % 100) == 0)
	myprintf(_CR "Free %d objs out of %d", nb_free, nb_alloc);

      SOS_ASSERT_FATAL(nb_free <= nb_alloc);
    }
  myprintf(_CR "Allocated %d objs, free %d objs", nb_alloc, nb_free);

  SOS_ASSERT_FATAL(nb_alloc == nb_free);

  SOS_ASSERT_FATAL(sos_kmem_cache_destroy(cache) == SOS_OK);
}


static void display_pmm(char *str)
{
  sos_count_t total, used;
  sos_physmem_get_state(& total, & used);
  sos_bochs_printf("PM %s: total=%d used=%d\n", str, total, used);
}

int MALLOC_TEST(void) {
  int i;

  display_pmm("Begin custom cache test");
  myprintf(_CR "custom cache test...");
  kmem_cache_test();

  display_pmm("Begin mm campaign");

  myprintf("Init alloc_table...\n");
  for (i = 0 ; i < NB_ALLOC_MAX ; i++) {
    _yield();
    alloc_table[i].size      = -1;
    alloc_table[i].base_addr = 0;
  }

  sos_kmem_vmm_dump_vr("Begin Kmem test campaign");
  for (i = 0 ; i < NB_CAMPAIGNS ; i++) {
    _yield();
    myprintf(_CR "----- Test campaign %d/%d ----\n", i+1, NB_CAMPAIGNS);
    if (test_campaign(rnd_int1(NB_ALLOC_MAX+1))) {
      myprintf("\nError\n");
      return -1;
    }
  }
  myprintf(_CR "Max allocated Mem = %d, Remaining alloc size = %d\n",
	   max_alloc_size, current_alloc_size);

  display_pmm("mm campaign: before final free");

  myprintf(_CR "---------------------------\n# Final Free...\n");
  for (i = 0 ; i < NB_ALLOC_MAX ; i++) {
    _yield();
    if (! alloc_table[i].base_addr)
      continue;
#ifdef CHECK_FULL_AREAS
      {
	int j;
	for (j = 0 ; j*sizeof(int) < alloc_table[i].size ; j++)
	  if (((int*)alloc_table[i].base_addr)[j] != i)
#else
	if (*((int*)alloc_table[i].base_addr) != i)
#endif
	  {
	    myprintf("alloc[%d]=0x%x - Wrong magic\n", i,
		     alloc_table[i].base_addr);
	    continue;
	  }
#ifdef CHECK_FULL_AREAS
      }
#endif
    kfree((void*)alloc_table[i].base_addr);
    current_alloc_size -= alloc_table[i].size;
    alloc_table[i].size      = -1;
    alloc_table[i].base_addr = 0;
  }

  SOS_ASSERT_FATAL(current_alloc_size == 0);

  display_pmm("mm campaign: the end");
  myprintf("The end.\n");
  sos_kmem_vmm_dump_vr("The end tests campaign");
  return 0;
}
// [[[]]]
